import logging
from typing import List, Union

import numpy as np

from ada import FEM, Assembly, Beam, Node, Part, Plate
from ada.config import Settings
from ada.core.utils import vector_length
from ada.fem import Bc, Connector, ConnectorSection, Constraint, Elem, FemSet

from .shapes import ElemShape


def get_eldata(fem_source: Union[Assembly, Part, FEM]):
    """Return a dictionary of basic mesh statistics"""

    el_types = dict()

    def scan_elem(mesh):
        for el in mesh.elements:
            if el.type not in el_types.keys():
                el_types[el.type] = 1
            else:
                el_types[el.type] += 1

    if type(fem_source) is Assembly:
        for p in fem_source.parts.values():
            scan_elem(p.fem)
    elif issubclass(type(fem_source), Part):
        scan_elem(fem_source.fem)
    elif type(fem_source) is FEM:
        scan_elem(fem_source)
    else:
        raise ValueError(f'Unknown fem_source "{fem_source}"')
    return el_types


def convert_springs_to_connectors(assembly: Assembly):
    """Converts all single noded springs to connector elements"""
    for p in assembly.get_all_subparts():
        for spring in p.fem.springs.values():
            n1 = spring.nodes[0]
            n2 = Node(n1.p - np.array([0, 0, 10e-3]))
            assembly.fem.add_rp(spring.name + "_rp", n2)
            fs = FemSet(spring.name + "_bc", [n2], "nset")
            assembly.fem.add_set(fs)
            assembly.fem.add_bc(Bc(spring.name + "_bc", fs, [1, 2, 3, 4, 5, 6]))
            diag = []
            for dof, row in enumerate(spring.stiff):
                for j, stiffness in enumerate(row):
                    if dof == j:
                        diag.append(stiffness)

            con_sec = ConnectorSection(spring.name + "_consec", diag, [])
            assembly.fem.add_connector_section(con_sec)
            con = Connector(spring.name + "_con", spring.id, n1, n2, "bushing", con_sec)
            assembly.fem.add_connector(con)
        p.fem._springs = dict()
        p.fem.elements.filter_elements(delete_elem=["SPRING1"])


def get_beam_end_nodes(bm: Beam, end=1, tol=1e-3) -> List[Node]:
    """Get list of nodes from end of beam"""
    p = bm.parent
    nodes = p.fem.nodes
    w = bm.section.w_btn
    h = bm.section.h
    xv = np.array(bm.xvec)
    yv = np.array(bm.yvec)
    zv = np.array(bm.up)
    if end == 1:
        p = bm.n1.p
    else:
        p = bm.n2.p
    n_min = p - xv * tol - (h / 2 + tol) * zv - (w / 2 + tol) * yv
    n_max = p + xv * tol + (h / 2 + tol) * zv + (w / 2 + tol) * yv
    members = [e for e in nodes.get_by_volume(n_min, n_max)]
    return members


def get_nodes_along_plate_edges(pl: Plate, fem: FEM, edge_indices=None, tol=1e-3) -> List[Node]:
    """Return FEM nodes from edges of a plate"""

    res = []
    bmin, bmax = list(zip(*pl.bbox))
    bmin_smaller = np.array(bmin) + pl.poly.xdir * tol + pl.poly.ydir * tol
    bmax_smaller = np.array(bmax) - pl.poly.xdir * tol - pl.poly.ydir * tol
    all_res = fem.nodes.get_by_volume(bmin, bmax)
    res = [n.id for n in fem.nodes.get_by_volume(bmin_smaller, bmax_smaller)]
    return list(filter(lambda x: x.id not in res, all_res))


def is_line_elem(elem: Elem):

    return True if elem.type in ElemShape.TYPES.lines else False


def convert_ecc_to_mpc(fem: FEM):
    """Converts beam offsets to MPC constraints"""
    edited_nodes = dict()
    tol = Settings.point_tol

    def build_constraint(n_old, elem, ecc, i):
        n_new = edited_nodes[n_old.id]
        mat = np.eye(3)
        new_p = np.dot(mat, ecc) + n_old.p
        n_new_ = Node(new_p, parent=elem.parent)
        if vector_length(n_new_.p - n_new.p) > tol:
            elem.parent.nodes.add(n_new_, allow_coincident=True)
            m_set = FemSet(f"el{elem.id}_mpc{i + 1}_m", [n_new_], "nset")
            s_set = FemSet(f"el{elem.id}_mpc{i + 1}_s", [n_old], "nset")
            c = Constraint(
                f"el{elem.id}_mpc{i + 1}_co",
                "mpc",
                m_set,
                s_set,
                mpc_type="Beam",
                parent=elem.parent,
            )
            elem.parent.add_constraint(c)
            elem.nodes[i] = n_new_
            edited_nodes[n_old.id] = n_new_

        else:
            elem.nodes[i] = n_new
            edited_nodes[n_old.id] = n_new

    def build_mpc_for_end(elem, n_old, ecc, i):
        if n_old.id in edited_nodes.keys():
            build_constraint(n_old, elem, ecc, i)
        else:
            mat = np.eye(3)
            new_p = np.dot(mat, ecc) + n_old.p
            n_new = Node(new_p, parent=elem.parent)
            elem.parent.nodes.add(n_new, allow_coincident=True)
            m_set = FemSet(f"el{elem.id}_mpc{i + 1}_m", [n_new], "nset")
            s_set = FemSet(f"el{elem.id}_mpc{i + 1}_s", [n_old], "nset")
            c = Constraint(
                f"el{elem.id}_mpc{i + 1}_co",
                "mpc",
                m_set,
                s_set,
                mpc_type="Beam",
                parent=elem.parent,
            )
            elem.parent.add_constraint(c)
            elem.nodes[i] = n_new
            edited_nodes[n_old.id] = n_new

    def build_mpc(elem: Elem):
        if elem.eccentricity.end1 is not None:
            n_old = elem.eccentricity.end1.node
            ecc = elem.eccentricity.end1.ecc_vector
            i = elem.nodes.index(n_old)
            build_mpc_for_end(elem, n_old, ecc, i)
        if elem.eccentricity.end2 is not None:
            n_old = elem.eccentricity.end2.node
            ecc = elem.eccentricity.end2.ecc_vector
            i = elem.nodes.index(n_old)
            build_mpc_for_end(elem, n_old, ecc, i)

    [build_mpc(el) for el in fem.elements.lines_ecc]


def convert_hinges_2_couplings(fem: FEM):
    """Convert beam hinges to coupling constraints"""
    from ada.core.utils import Counter
    from ada.fem.elements import Hinge

    constrain_ids = []

    max_node_id = fem.nodes.max_nid
    new_node_id = Counter(max_node_id + 10000)

    def convert_hinge(elem: Elem, hinge: Hinge):
        if hinge.constraint_ref is not None:
            return
        n = hinge.fem_node
        csys = hinge.csys
        d = hinge.retained_dofs

        n2 = Node(n.p, next(new_node_id), parent=elem.parent)
        elem.parent.nodes.add(n2, allow_coincident=True)
        i = elem.nodes.index(n)
        elem.nodes[i] = n2

        if elem.eccentricity is not None:
            if elem.eccentricity.end1 is not None:
                if n == elem.eccentricity.end1.node:
                    elem.eccentricity.end1.node = n2

            if elem.eccentricity.end2 is not None:
                if n == elem.eccentricity.end2.node:
                    elem.eccentricity.end2.node = n2

        if n2.id not in constrain_ids:
            constrain_ids.append(n2.id)
        else:
            logging.error(f"Hinged node {n2} cannot be added twice to different couplings")
            return None

        m_set = FemSet(f"el{elem.id}_hinge{i + 1}_m", [n], "nset")
        s_set = FemSet(f"el{elem.id}_hinge{i + 1}_s", [n2], "nset")

        elem.parent.add_set(m_set)
        elem.parent.add_set(s_set)
        c = Constraint(
            f"el{elem.id}_hinge{i + 1}_co",
            "coupling",
            m_set,
            s_set,
            d,
            csys=csys,
        )
        elem.parent.add_constraint(c)
        hinge.constraint_ref = c
        logging.info(f"added constraint {c}")

    for el in fem.elements.lines_hinged:
        if el.hinge_prop.end1 is not None:
            convert_hinge(el, el.hinge_prop.end1)
        if el.hinge_prop.end2 is not None:
            convert_hinge(el, el.hinge_prop.end2)


def is_tri6_shell_elem(sh_fs):
    elem_check = [x.type in ElemShape.TYPES.tri6 for x in sh_fs.elset.members]
    return all(elem_check)


def is_quad8_shell_elem(sh_fs):
    elem_check = [x.type in ElemShape.TYPES.quad8 for x in sh_fs.elset.members]
    return all(elem_check)



from __future__ import annotations

import pathlib
from typing import Union

from ada import fem
from ada.concepts.connections import Bolts, Weld
from ada.concepts.curves import ArcSegment, CurvePoly, CurveRevolve, LineSegment
from ada.concepts.levels import FEM, Assembly, Part
from ada.concepts.piping import Pipe, PipeSegElbow, PipeSegStraight
from ada.concepts.points import Node
from ada.concepts.primitives import (
    Penetration,
    PrimBox,
    PrimCyl,
    PrimExtrude,
    PrimRevolve,
    PrimSphere,
    PrimSweep,
    Shape,
)
from ada.concepts.structural import Beam, Plate, Wall
from ada.concepts.transforms import Placement
from ada.config import User
from ada.materials import Material
from ada.sections import Section

__author__ = "Kristoffer H. Andersen"


def from_ifc(ifc_file: Union[str, pathlib.Path]) -> Assembly:
    a = Assembly()
    a.read_ifc(ifc_file)
    return a


def from_step(step_file: Union[str, pathlib.Path], source_units="m", **kwargs) -> Assembly:
    a = Assembly()
    a.read_step_file(step_file, source_units=source_units, **kwargs)
    return a


def from_fem(
    fem_file: Union[str, list, pathlib.Path],
    fem_format: Union[str, list] = None,
    name: Union[str, list] = None,
    enable_experimental_cache=False,
    source_units="m",
    fem_converter="default",
) -> Assembly:
    a = Assembly(enable_experimental_cache=enable_experimental_cache, units=source_units)
    if type(fem_file) is str or issubclass(type(fem_file), pathlib.Path):
        a.read_fem(fem_file, fem_format, name, fem_converter=fem_converter)
    elif type(fem_file) is list:
        for i, f in enumerate(fem_file):
            fem_format_in = fem_format if fem_format is None else fem_format[i]
            name_in = name if name is None else name[i]
            a.read_fem(f, fem_format_in, name_in, fem_converter=fem_converter)
    else:
        raise ValueError(f'fem_file must be either string or list. Passed type was "{type(fem_file)}"')

    return a


__all__ = [
    "Assembly",
    "Part",
    "FEM",
    "from_ifc",
    "from_fem",
    "Beam",
    "Plate",
    "Pipe",
    "PipeSegStraight",
    "PipeSegElbow",
    "Wall",
    "Penetration",
    "Section",
    "Material",
    "Shape",
    "Node",
    "Placement",
    "PrimBox",
    "PrimCyl",
    "PrimExtrude",
    "PrimRevolve",
    "PrimSphere",
    "PrimSweep",
    "CurvePoly",
    "CurveRevolve",
    "LineSegment",
    "ArcSegment",
    "User",
    "Bolts",
    "Weld",
    "fem",
]



from __future__ import annotations

import logging
import os
import pathlib
import subprocess
from dataclasses import dataclass, field
from typing import List, Tuple, Union

import meshio
import numpy as np
import pythreejs
from IPython.display import display
from ipywidgets import Dropdown, HBox, VBox

from ada.fem.formats import FEATypes
from ada.visualize.femviz import get_edges_and_faces_from_meshio, magnitude
from ada.visualize.renderer_pythreejs import MyRenderer
from ada.visualize.threejs_utils import edges_to_mesh, faces_to_mesh, vertices_to_mesh

from .concepts.eigenvalue import EigenDataSummary


class Results:
    def __init__(self, res_path, name=None, fem_format=None, assembly=None, palette=None, output=None, overwrite=True):
        self._name = name
        self._visualizer = ResultsMesh(palette, fem_format=fem_format, parent=self)
        self._eigen_mode_data = None
        self._fem_format = fem_format
        self._assembly = assembly
        self._output = output
        self._overwrite = overwrite
        self._results_file_path = pathlib.Path(res_path)
        self._read_result_file(self.results_file_path)

    def _read_result_file(self, file_ref, overwrite=False):
        if file_ref.exists() is False:
            return None

        mesh = self._get_results_from_result_file(file_ref)

        if mesh is None:
            return None

        print(f'Importing meshio.Mesh from result file "{file_ref}"')
        self.result_mesh.add_results(mesh)

    def _get_results_from_result_file(self, file_ref, overwrite=False):
        from .formats.abaqus.results import read_abaqus_results
        from .formats.calculix.results import read_calculix_results
        from .formats.code_aster.results import read_code_aster_results
        from .formats.sesam.results import read_sesam_results

        file_ref = pathlib.Path(file_ref)
        suffix = file_ref.suffix.lower()

        res_map = {
            ".rmed": (read_code_aster_results, FEATypes.CODE_ASTER),
            ".frd": (read_calculix_results, FEATypes.CALCULIX),
            ".odb": (read_abaqus_results, FEATypes.ABAQUS),
            ".sin": (read_sesam_results, FEATypes.SESAM),
        }
        res_reader, fem_format = res_map.get(suffix, (None, None))

        if res_reader is None:
            logging.error(f'Results class currently does not support filetype "{suffix}"')
            return None

        self.fem_format = fem_format

        return res_reader(self, file_ref, overwrite)

    def save_output(self, dest_file) -> None:
        if self.output is None or self.output.stdout is None:
            print("No output is found")
            return None
        dest_file = pathlib.Path(dest_file)

        os.makedirs(dest_file.parent, exist_ok=True)
        with open(dest_file, "w") as f:
            f.write(self.output.stdout)

    @property
    def name(self):
        return self._name

    @property
    def fem_format(self):
        return self._fem_format

    @fem_format.setter
    def fem_format(self, value):
        if value not in FEATypes.all:
            raise ValueError(f'Unsupported FEA Type "{value}"')
        self._fem_format = value

    @property
    def output(self) -> subprocess.CompletedProcess:
        return self._output

    @property
    def results_file_path(self):
        return self._results_file_path

    @results_file_path.setter
    def results_file_path(self, value):
        self._results_file_path = value

    @property
    def result_mesh(self) -> ResultsMesh:
        return self._visualizer

    @property
    def assembly(self):
        """:rtype: ada.Assembly"""
        return self._assembly

    @property
    def eigen_mode_data(self) -> EigenDataSummary:
        return self._eigen_mode_data

    @eigen_mode_data.setter
    def eigen_mode_data(self, value: EigenDataSummary):
        self._eigen_mode_data = value

    def _repr_html_(self):

        if self.result_mesh.renderer is None:
            res = self.result_mesh.build_renderer()
        else:
            res = True

        if res is False:
            return

        p3s_renderer = self.result_mesh.renderer
        display(HBox([VBox([HBox(p3s_renderer.controls), p3s_renderer.renderer]), p3s_renderer.html]))

    def __repr__(self):
        return f"Results({self._fem_format}, {self._results_file_path.name})"


@dataclass
class ResultsMesh:

    palette: List[tuple]
    parent: Results
    fem_format: str
    renderer: MyRenderer = None
    render_sets: Dropdown = None
    mesh: meshio.Mesh = None
    undeformed_mesh: Union[None, pythreejs.Mesh] = None
    deformed_mesh: Tuple[pythreejs.Mesh, pythreejs.Points, pythreejs.LineSegments] = None
    point_data: List = field(default_factory=list)
    cell_data: List = field(default_factory=list)

    vertices: np.ndarray = None
    edges: np.ndarray = None
    faces: np.ndarray = None

    def __post_init__(self):
        self.palette = [(0, 149 / 255, 239 / 255), (1, 0, 0)] if self.palette is None else self.palette

    def add_results(self, mesh: meshio.Mesh):
        self.mesh = mesh
        self.vertices = np.asarray(mesh.points, dtype="float32")

        edges, faces = get_edges_and_faces_from_meshio(mesh)
        self.edges = np.asarray(edges, dtype="uint16").ravel()
        self.faces = np.asarray(faces, dtype="uint16").ravel()

        for n in mesh.point_data.keys():
            self.point_data.append(n)

        for n in mesh.cell_data.keys():
            self.cell_data.append(n)

    def build_renderer(self) -> bool:
        self.renderer = MyRenderer()
        if len(self.point_data) == 0:
            return False
        if self.fem_format == FEATypes.CODE_ASTER:
            data = [x for x in self.point_data if "DISP" in x][-1]
        elif self.fem_format == FEATypes.CALCULIX:
            data = [x for x in self.point_data if "U" in x][-1]
        else:
            raise NotImplementedError(f'Support for analysis_type "{self.fem_format}"')

        self.create_viz_geom(data, displ_data=True, renderer=self.renderer)
        i = self.point_data.index(data)
        self.render_sets = Dropdown(
            options=self.point_data, value=self.point_data[i], tooltip="Select a set", disabled=False
        )
        self.render_sets.observe(self.on_changed_point_data_set, "value")
        self.renderer.controls.pop()
        self.renderer.controls.append(self.render_sets)
        return True

    def _colorize_data(self, data, func=magnitude):
        res = [func(d) for d in data]
        sorte = sorted(res)
        min_r = sorte[0]
        max_r = sorte[-1]

        start = np.array(self.palette[0])
        end = np.array(self.palette[-1])

        def curr_p(t):
            return start + (end - start) * t / (max_r - min_r)

        colors = np.asarray([curr_p(x) for x in res], dtype="float32")
        return colors

    def create_viz_geom(self, data_type, displ_data=False, renderer: MyRenderer = None) -> None:
        default_vertex_color = (8, 8, 8)

        data = np.asarray(self.mesh.point_data[data_type], dtype="float32")
        colors = self._colorize_data(data)

        if renderer is None:
            renderer = MyRenderer()
            self.renderer = renderer

        # deformations
        if displ_data is True:
            vertices = np.asarray([x + u[:3] for x, u in zip(self.vertices, data)], dtype="float32")
            if self.undeformed_mesh is None:
                dark_grey = (0.66, 0.66, 0.66)
                white_color = np.asarray([dark_grey for x in self.vertices], dtype="float32")
                o_mesh = faces_to_mesh("undeformed", self.vertices, self.faces, white_color, opacity=0.5)
                self.undeformed_mesh = o_mesh
                renderer._displayed_non_pickable_objects.add(o_mesh)
        else:
            vertices = self.vertices
            if self.undeformed_mesh is not None:
                renderer._displayed_non_pickable_objects.remove(self.undeformed_mesh)
                self.undeformed_mesh = None

        vertices = np.array(vertices, dtype=np.float32)

        # Colours
        mesh = faces_to_mesh("deformed", vertices, self.faces, colors)
        points = vertices_to_mesh("deformed_vertices", vertices, default_vertex_color)
        lines = edges_to_mesh("deformed_lines", vertices, self.edges, default_vertex_color)

        if self.deformed_mesh is None:
            self.deformed_mesh = (mesh, points, lines)
            renderer.displayed_pickable_objects.add(mesh)
            renderer.displayed_pickable_objects.add(points)
            renderer.displayed_pickable_objects.add(lines)
            renderer.build_display(camera_type="perspective")
        else:
            face_geom = self.deformed_mesh[0].geometry
            face_geom.attributes["position"].array = vertices
            face_geom.attributes["index"].array = self.faces
            face_geom.attributes["color"].array = colors

            point_geom = self.deformed_mesh[1].geometry
            point_geom.attributes["position"].array = vertices

            edge_geom = self.deformed_mesh[2].geometry
            edge_geom.attributes["position"].array = vertices
            edge_geom.attributes["index"].array = self.edges

    def on_changed_point_data_set(self, p):
        data = p["new"]
        if self.fem_format == FEATypes.CODE_ASTER:
            is_displ = True if "DISP" in data else False
            if "point_tags" in data:
                print("\r" + "Point Tags are not a valid display value" + 10 * " ", end="")
                return None
        elif self.fem_format == FEATypes.CALCULIX:
            is_displ = True if "U" in data else False
        else:
            return None

        if is_displ:
            self.create_viz_geom(data, displ_data=True, renderer=self.renderer)
        else:
            self.create_viz_geom(data, renderer=self.renderer)


def get_fem_stats(fem_file, dest_md_file, data_file="data.json"):
    """

    :param fem_file:
    :param dest_md_file:
    :param data_file: Destination of data.json file (keeping track of last modified status etc..)
    """
    import json
    import os

    from ada import Assembly
    from ada.fem.utils import get_eldata

    dest_md_file = pathlib.Path(dest_md_file)
    data_file = pathlib.Path(data_file)
    a = Assembly()
    a.read_fem(fem_file)

    out_str = ""

    for name, part in a.parts.items():
        fem = part.fem
        r = get_eldata(fem_source=fem)
        if len(r.keys()) == 0:
            continue
        out_str += f"* **{name}**: ("

        el_data = ""
        for el_type, el_num in r.items():
            el_data += f"{el_type}: {el_num}, "

        out_str += el_data[:-2] + ")\n"

    os.makedirs(dest_md_file.parent, exist_ok=True)

    with open(dest_md_file, "w") as f:
        f.write(out_str)

    if data_file.exists():
        with open(data_file, "r") as f:
            data = json.load(f)
    else:
        data = dict()
    print(data)



import datetime
import logging
from operator import attrgetter

from ada.concepts.levels import FEM, Part
from ada.core.utils import Counter, get_current_user

from .templates import top_level_fem_str
from .write_utils import write_ff


def to_fem(assembly, name, analysis_dir=None, metadata=None):
    from .write_constraints import constraint_str
    from .write_elements import elem_str
    from .write_loads import loads_str
    from .write_masses import mass_str
    from .write_sections import sections_str
    from .write_steps import write_sestra_inp

    if metadata is None:
        metadata = dict()

    if "control_file" not in metadata.keys():
        metadata["control_file"] = None

    parts = list(filter(lambda x: len(x.fem.nodes) > 0, assembly.get_all_subparts()))
    if len(parts) != 1:
        raise ValueError("Sesam writer currently only works for a single part")

    if len(assembly.fem.steps) > 1:
        logging.error("Sesam writer currently only supports 1 step. Will only use 1st step")

    part = parts[0]

    thick_map = dict()

    now = datetime.datetime.now()
    date_str = now.strftime("%d-%b-%Y")
    clock_str = now.strftime("%H:%M:%S")
    user = get_current_user()

    units = "UNITS     5.00000000E+00  1.00000000E+00  1.00000000E+00  1.00000000E+00\n          1.00000000E+00\n"

    inp_file_path = (analysis_dir / f"{name}T1").with_suffix(".FEM")
    if len(assembly.fem.steps) > 0:
        step = assembly.fem.steps[0]
        with open(analysis_dir / "sestra.inp", "w") as f:
            f.write(write_sestra_inp(name, step))

    with open(inp_file_path, "w") as d:
        d.write(top_level_fem_str.format(date_str=date_str, clock_str=clock_str, user=user))
        d.write(units)
        d.write(materials_str(part))
        d.write(sections_str(part.fem, thick_map))
        d.write(univec_str(part.fem))
        d.write(nodes_str(part.fem))
        d.write(mass_str(part.fem))
        d.write(bc_str(part.fem) + bc_str(assembly.fem))
        d.write(constraint_str(part.fem) + constraint_str(assembly.fem))
        d.write(hinges_str(part.fem))
        d.write(elem_str(part.fem, thick_map))
        d.write(loads_str(assembly.fem) + loads_str(part.fem))
        d.write("IEND                0.00            0.00            0.00            0.00")

    print(f'Created an Sesam input deck at "{analysis_dir}"')


def materials_str(part: Part):
    out_str = "".join(
        [write_ff("TDMATER", [(4, mat.id, 100 + len(mat.name), 0), (mat.name,)]) for mat in part.materials]
    )

    out_str += "".join(
        [
            write_ff(
                "MISOSEL",
                [
                    (mat.id, mat.model.E, mat.model.v, mat.model.rho),
                    (mat.model.zeta, mat.model.alpha, 1, mat.model.sig_y),
                ],
            )
            for mat in part.materials
        ]
    )
    return out_str


def nodes_str(fem: FEM) -> str:
    nodes = sorted(fem.nodes, key=attrgetter("id"))

    nids = []
    for n in nodes:
        if n.id not in nids:
            nids.append(n.id)
        else:
            raise Exception('Doubly defined node id "{}". TODO: Make necessary code updates'.format(n[0]))
    if len(nodes) == 0:
        return "** No Nodes"
    else:

        out_str = "".join([write_ff("GNODE", [(no.id, no.id, 6, 123456)]) for no in nodes])
        out_str += "".join([write_ff("GCOORD", [(no.id, no[0], no[1], no[2])]) for no in nodes])
        return out_str


def bc_str(fem: FEM) -> str:
    out_str = ""
    for bc in fem.bcs:
        for m in bc.fem_set.members:
            dofs = [1 if i in bc.dofs else 0 for i in range(1, 7)]
            data = [tuple([m.id, 6] + dofs[:2]), tuple(dofs[2:])]
            out_str += write_ff("BNBCD", data)
    return out_str


def hinges_str(fem: FEM) -> str:
    out_str = ""
    h = Counter(1)

    def write_hinge(hinge):
        dofs = [0 if i in hinge else 1 for i in range(1, 7)]
        fix_id = next(h)
        data = [tuple([fix_id, 3, 0, 0]), tuple(dofs[:4]), tuple(dofs[4:])]
        return fix_id, write_ff("BELFIX", data)

    for el in fem.elements:
        h1, h2 = el.metadata.get("h1", None), el.metadata.get("h2", None)
        if h2 is None and h1 is None:
            continue
        h1_fix, h2_fix = 0, 0
        if h1 is not None:
            h1_fix, res_str = write_hinge(h1)
            out_str += res_str
        if h2 is not None:
            h2_fix, res_str = write_hinge(h2)
            out_str += res_str
        el.metadata["fixno"] = h1_fix, h2_fix

    return out_str


def univec_str(fem: FEM) -> str:
    out_str = ""
    uvec_id = Counter(1)

    unit_vecs = dict()

    def write_local_z(vec):
        tvec = tuple(vec)
        if tvec in unit_vecs.keys():
            return unit_vecs[tvec], None
        trans_no = next(uvec_id)
        data = [tuple([trans_no, *vec])]
        unit_vecs[tvec] = trans_no
        return trans_no, write_ff("GUNIVEC", data)

    for el in fem.elements:
        local_z = el.fem_sec.local_z
        transno, res_str = write_local_z(local_z)
        if res_str is None:
            el.metadata["transno"] = transno
            continue
        out_str += res_str
        el.metadata["transno"] = transno

    return out_str



import logging

from ada.concepts.levels import Assembly, Part
from ada.fem.formats.ifc.writer import to_ifc_fem

from .utils import create_guid


def add_part_objects_to_ifc(p: Part, f, assembly: Assembly, ifc_include_fem=False):
    # TODO: Consider having all of these operations happen upon import of elements as opposed to one big operation
    #  on export

    part_ifc = p.get_ifc_elem()
    owner_history = assembly.user.to_ifc()
    physical_objects = []
    for m in p.materials.name_map.values():
        f.add(m.ifc_mat)

    for bm in p.beams:
        bm_ifc = bm.get_ifc_elem()
        f.add(bm_ifc)
        physical_objects.append(bm_ifc)

    for pl in p.plates:
        pl_ifc = pl.get_ifc_elem()
        f.add(pl_ifc)
        physical_objects.append(pl_ifc)

    for pi in p.pipes:
        logging.debug(f'Creating IFC Elem for PIPE "{pi.name}"')
        f.add(pi.get_ifc_elem())

    for wall in p.walls:
        f.add(wall.get_ifc_elem())
        physical_objects.append(wall.get_ifc_elem())

    for shp in p.shapes:
        if "ifc_file" in shp.metadata.keys():
            ifc_file = shp.metadata["ifc_file"]
            ifc_f = assembly.get_ifc_source_by_name(ifc_file)
            ifc_elem = ifc_f.by_guid(shp.guid)
            f.add(ifc_elem)
            physical_objects.append(ifc_elem)
        else:
            f.add(shp.get_ifc_elem())
            physical_objects.append(shp.get_ifc_elem())

    if len(p.fem.nodes) > 0 and ifc_include_fem is True:
        to_ifc_fem(p.fem, f)

    if len(physical_objects) == 0:
        return

    f.create_entity(
        "IfcRelContainedInSpatialStructure",
        create_guid(),
        owner_history,
        "Physical model",
        None,
        physical_objects,
        part_ifc,
    )



import logging

import numpy as np

from ada import Section
from ada.config import Settings
from ada.sections.categories import SectionCat

from .utils import create_guid, create_ifcindexpolyline, create_ifcpolyline


def export_beam_section(section: Section):
    a = section.parent.parent.get_assembly()
    f = a.ifc_file
    sec_props = dict(ProfileType="AREA", ProfileName=section.name)
    section_profile = section.get_section_profile(True)
    if SectionCat.is_i_profile(section.type):
        if Settings.use_param_profiles is False:
            polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)

            ifc_sec_type = "IfcArbitraryClosedProfileDef"
            sec_props.update(dict(OuterCurve=polyline))
        else:
            if SectionCat.is_strong_axis_symmetric(section) is False:
                logging.error(
                    "Note! Not using IfcAsymmetricIShapeProfileDef as it is not supported by ifcopenshell v IFC4"
                )
                # ifc_sec_type = "IfcAsymmetricIShapeProfileDef"
                # sec_props.update(
                #     dict(
                #         TopFlangeWidth=self.w_top,
                #         BottomFlangeWidth=self.w_btn,
                #         OverallDepth=self.h,
                #         WebThickness=self.t_w,
                #         TopFlangeThickness=self.t_ftop,
                #         BottomFlangeThickness=self.t_fbtn,
                #     )
                # )

            ifc_sec_type = "IfcIShapeProfileDef"
            sec_props.update(
                dict(
                    OverallWidth=section.w_top,
                    OverallDepth=section.h,
                    WebThickness=section.t_w,
                    FlangeThickness=section.t_ftop,
                )
            )

    elif SectionCat.is_angular(section.type):
        points = [f.createIfcCartesianPoint(p) for p in section_profile.outer_curve.points2d]
        ifc_polyline = f.createIfcPolyLine(points)
        ifc_sec_type = "IfcArbitraryClosedProfileDef"
        sec_props.update(dict(OuterCurve=ifc_polyline))

        if Settings.use_param_profiles is True:
            logging.debug(f'Export of "{section.type}" profile to parametric IFC profile is not yet added')

    elif SectionCat.is_box_profile(section.type):
        ot_disc = section_profile.outer_curve.points2d
        in_disc = section_profile.inner_curve.points2d
        outer_points = [f.createIfcCartesianPoint(p) for p in ot_disc + [ot_disc[0]]]
        inner_points = [f.createIfcCartesianPoint(p) for p in in_disc + [in_disc[0]]]
        inner_curve = f.createIfcPolyLine(inner_points)
        outer_curve = f.createIfcPolyLine(outer_points)
        ifc_sec_type = "IfcArbitraryProfileDefWithVoids"
        sec_props.update(dict(OuterCurve=outer_curve, InnerCurves=[inner_curve]))

        if Settings.use_param_profiles is True:
            logging.debug(f'Export of "{section.type}" profile to parametric IFC profile is not yet added')

    elif section.type in SectionCat.circular:
        ifc_sec_type = "IfcCircleProfileDef"
        sec_props.update(dict(Radius=section.r))
    elif section.type in SectionCat.tubular:
        ifc_sec_type = "IfcCircleHollowProfileDef"
        sec_props.update(dict(Radius=section.r, WallThickness=section.wt))
    elif section.type in SectionCat.general:
        logging.error("Note! Creating a Circle profile from general section (just for visual inspection as of now)")
        r = np.sqrt(section.properties.Ax / np.pi)
        ifc_sec_type = "IfcCircleProfileDef"
        sec_props.update(dict(Radius=r))
    elif section.type in SectionCat.flatbar:
        polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)
        ifc_sec_type = "IfcArbitraryClosedProfileDef"
        sec_props.update(dict(OuterCurve=polyline))

        if Settings.use_param_profiles is True:
            logging.debug(f'Export of "{section.type}" profile to parametric IFC profile is not yet added')

    elif section.type in SectionCat.channels:
        if Settings.use_param_profiles is False:
            polyline = create_ifcpolyline(f, section_profile.outer_curve.points2d)
            ifc_sec_type = "IfcArbitraryClosedProfileDef"
            sec_props.update(dict(OuterCurve=polyline))
        else:
            ifc_sec_type = "IfcUShapeProfileDef"
            sec_props.update(
                dict(
                    Depth=section.h, FlangeWidth=section.w_top, WebThickness=section.t_w, FlangeThickness=section.t_ftop
                )
            )
    elif section.type == "poly":
        opoly = section.poly_outer
        opoints = [(float(n[0]), float(n[1]), float(n[2])) for n in opoly.seg_global_points]
        opolyline = create_ifcindexpolyline(f, opoints, opoly.seg_index)
        if section.poly_inner is None:
            ifc_sec_type = "IfcArbitraryClosedProfileDef"
            sec_props.update(dict(OuterCurve=opolyline))
        else:
            ipoly = section.poly_inner
            ipoints = [(float(n[0]), float(n[1]), float(n[2])) for n in ipoly.seg_global_points]
            ipolyline = create_ifcindexpolyline(f, ipoints, ipoly.seg_index)
            ifc_sec_type = "IfcArbitraryProfileDefWithVoids"
            sec_props.update(dict(OuterCurve=opolyline, InnerCurves=[ipolyline]))
    else:
        raise ValueError(f'Have yet to implement section type "{section.type}"')

    if section.name is None:
        raise ValueError("Name cannot be None!")

    profile = f.create_entity(ifc_sec_type, **sec_props)

    beamtype = f.create_entity(
        "IfcBeamType",
        create_guid(),
        a.user.to_ifc(),
        section.name,
        section.sec_str,
        None,
        None,
        None,
        None,
        None,
        "BEAM",
    )
    return profile, beamtype



gmsh_map = {
    "Triangle 3": "S3",
    "Triangle 6": "STRI65",
    "Quadrilateral 4": "S4R",
    "Quadrilateral 8": "S8R",
    "Line 2": "B31",
    "Line 3": "B32",
    "Tetrahedron 4": "C3D4",
    "Tetrahedron 10": "C3D10",
}



def equation_compiler(f, print_latex=False, print_formula=False):
    from inspect import getsourcelines

    try:
        import pytexit
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "To use the equation compiler you will need to install pytexit first.\n"
            'Use "pip install pytexit"\n\n'
            f'Original error message: "{e}"'
        )

    lines = getsourcelines(f)
    final_line = lines[0][-1]
    return pytexit.py2tex(final_line.replace("return ", ""), print_latex=print_latex, print_formula=print_formula)



import logging

from ada import FEM
from ada.fem import FemSection
from ada.fem.steps import StepExplicit
from ada.sections import GeneralProperties, Section, SectionCat

log_fin = "Please check your result and input. This is not a validated method of solving this issue"


def section_str(fem: FEM):
    solids = fem.sections.solids
    shells = fem.sections.shells
    lines = fem.sections.lines

    solid_secs_str = "\n".join([solid_section_str(so) for so in solids]) if len(solids) > 0 else "** No solid sections"
    shell_secs_str = "\n".join([shell_section_str(sh) for sh in shells]) if len(shells) > 0 else "** No shell sections"
    line_secs_str = "\n".join([line_section_str(li) for li in lines]) if len(lines) > 0 else "** No line sections"

    return solid_secs_str.strip() + shell_secs_str.strip() + line_secs_str.strip()


def solid_section_str(fem_sec: FemSection):
    return f"""** Section: {fem_sec.name}
*Solid Section, elset={fem_sec.elset.name}, material={fem_sec.material.name}
,"""


def shell_section_str(fem_sec: FemSection):
    return f"""** Section: {fem_sec.name}
*Shell Section, elset={fem_sec.elset.name}, material={fem_sec.material.name}
 {fem_sec.thickness}, {fem_sec.int_points}"""


def line_section_str(fem_sec: FemSection):
    top_line = f"** Section: {fem_sec.elset.name}  Profile: {fem_sec.elset.name}"
    density = fem_sec.material.model.rho if fem_sec.material.model.rho > 0.0 else 1e-4
    ass = fem_sec.parent.parent.get_assembly()

    rotary_str = ""
    if len(ass.fem.steps) > 0:
        initial_step = ass.fem.steps[0]
        if type(initial_step) is StepExplicit:
            rotary_str = ", ROTARY INERTIA=ISOTROPIC"
    sec_data = line_cross_sec_type_str(fem_sec)
    sec_props = line_section_props(fem_sec)
    if sec_data != "GENERAL":
        return (
            f"{top_line}\n*Beam Section, elset={fem_sec.elset.name}, material={fem_sec.material.name}, "
            f"temperature={line_temperature_str(fem_sec)}, section={sec_data}{rotary_str}\n{sec_props}"
        )
    else:
        return f"""{top_line}
*Beam General Section, elset={fem_sec.elset.name}, section=GENERAL{rotary_str}, density={density}
 {sec_props}"""


def line_section_props(fem_sec: FemSection):
    n1 = ", ".join(str(x) for x in fem_sec.local_y)
    if "line1" in fem_sec.metadata.keys():
        return fem_sec.metadata["line1"] + f"\n{n1}"

    sec = fem_sec.section
    sec_data = line_cross_sec_type_str(fem_sec)

    if sec_data == "CIRC":
        return f"{sec.r}\n {n1}"
    elif sec_data == "I":
        if sec.t_fbtn + sec.t_w > min(sec.w_top, sec.w_btn):
            # TODO: Evaluate why this was here
            # new_width = sec.t_fbtn + sec.t_w + 5e-3
            # if sec.w_btn == min(sec.w_top, sec.w_btn):
            #     sec.w_btn = new_width
            # else:
            #     sec.w_top = new_width
            logging.info(f"For {fem_sec.name}: t_fbtn + t_w > min(w_top, w_btn). {log_fin}")
        return f"{sec.h / 2}, {sec.h}, {sec.w_btn}, {sec.w_top}, {sec.t_fbtn}, {sec.t_ftop}, {sec.t_w}\n {n1}"
    elif sec_data == "BOX":
        if sec.t_w * 2 > min(sec.w_top, sec.w_btn):
            raise ValueError("Web thickness cannot be larger than section width")
        return f"{sec.w_top}, {sec.h}, {sec.t_w}, {sec.t_ftop}, {sec.t_w}, {sec.t_fbtn}\n {n1}"
    elif sec_data == "GENERAL":
        mat = fem_sec.material.model
        gp = eval_general_properties(sec)
        return f"{gp.Ax}, {gp.Iy}, {gp.Iyz}, {gp.Iz}, {gp.Ix}\n {n1}\n {mat.E:.3E}, {mat.G},{mat.alpha:.2E}"
    elif sec_data == "PIPE":
        return f"{sec.r}, {sec.wt}\n {n1}"
    elif sec_data == "L":
        return f"{sec.w_btn}, {sec.h}, {sec.t_fbtn}, {sec.t_w}\n {n1}"
    elif sec_data == "RECT":
        return f"{sec.w_btn}, {sec.h}\n {n1}"
    else:
        raise NotImplementedError(f'section type "{sec.type}" is not added to Abaqus export yet')


def line_cross_sec_type_str(fem_sec: FemSection):
    if "section_type" in fem_sec.metadata.keys():
        return fem_sec.metadata["section_type"]
    sec_type = fem_sec.section.type
    from ada.sections.categories import BaseTypes

    bt = BaseTypes
    base_type = SectionCat.get_shape_type(fem_sec.section)
    sec_map = {
        bt.CIRCULAR: "CIRC",
        bt.IPROFILE: "I",
        bt.BOX: "BOX",
        bt.GENERAL: "GENERAL",
        bt.TUBULAR: "PIPE",
        bt.ANGULAR: "L",
        bt.CHANNEL: "GENERAL",
        bt.FLATBAR: "RECT",
    }
    sec_str = sec_map.get(base_type, None)
    if sec_str is None:
        raise Exception(f'Section type "{sec_type}" is not added to Abaqus beam export yet')

    if base_type in [bt.CHANNEL]:
        logging.error(f'Profile type "{sec_type}" is not supported by Abaqus. Using a General Section instead')

    return sec_str


def line_temperature_str(fem_sec: FemSection):
    _temperature = fem_sec.metadata["temperature"] if "temperature" in fem_sec.metadata.keys() else None
    return _temperature if _temperature is not None else "GRADIENT"


def eval_general_properties(section: Section) -> GeneralProperties:
    gp = section.properties
    name = gp.parent.parent.name
    if gp.Ix <= 0.0:
        gp.Ix = 1
        logging.error(f"Section {name} Ix <= 0.0. Changing to 2. {log_fin}")
    if gp.Iy <= 0.0:
        gp.Iy = 2
        logging.error(f"Section {name} Iy <= 0.0. Changing to 2. {log_fin}")
    if gp.Iz <= 0.0:
        gp.Iz = 2
        logging.error(f"Section {name} Iz <= 0.0. Changing to 2. {log_fin}")
    if gp.Iyz <= 0.0:
        gp.Iyz = (gp.Iy + gp.Iz) / 2
        logging.error(f"Section {name} Iyz <= 0.0. Changing to (Iy + Iz) / 2. {log_fin}")
    if gp.Iy * gp.Iz - gp.Iyz ** 2 < 0:
        old_y = str(gp.Iy)
        gp.Iy = 1.1 * (gp.Iy + (gp.Iyz ** 2) / gp.Iz)
        logging.error(
            f"Warning! Section {name}: I(11)*I(22)-I(12)**2 MUST BE POSITIVE. " f"Mod Iy={old_y} to {gp.Iy}. {log_fin}"
        )
    if (-(gp.Iy + gp.Iz) / 2 < gp.Iyz <= (gp.Iy + gp.Iz) / 2) is False:
        raise ValueError("Iyz must be between -(Iy+Iz)/2 and (Iy+Iz)/2")
    return gp



import logging
import os
import pathlib
from typing import Union

from OCC.Core.BRepAlgoAPI import BRepAlgoAPI_Fuse
from OCC.Core.IFSelect import IFSelect_RetError
from OCC.Core.Interface import Interface_Static_SetCVal
from OCC.Core.STEPConstruct import stepconstruct_FindEntity
from OCC.Core.STEPControl import STEPControl_AsIs, STEPControl_Writer
from OCC.Core.TCollection import TCollection_HAsciiString

from ada.base.physical_objects import BackendGeom
from ada.concepts.levels import Assembly, Part
from ada.concepts.piping import Pipe
from ada.concepts.primitives import Shape
from ada.concepts.structural import Beam, Plate, Wall
from ada.core.utils import Counter
from ada.fem.shapes import ElemType

# Reference: https://www.opencascade.com/doc/occt-7.4.0/overview/html/occt_user_guides__step.html#occt_step_3

shp_names = Counter(1, "shp")
valid_types = Union[BackendGeom, Beam, Plate, Wall, Part, Assembly, Shape, Pipe]


class StepExporter:
    def __init__(self, schema="AP242", assembly_mode=1):
        self.writer = STEPControl_Writer()
        fp = self.writer.WS().TransferWriter().FinderProcess()
        self.fp = fp

        Interface_Static_SetCVal("write.step.schema", schema)
        # Interface_Static_SetCVal('write.precision.val', '1e-5')
        Interface_Static_SetCVal("write.precision.mode", "1")
        Interface_Static_SetCVal("write.step.assembly", str(assembly_mode))

    def add_to_step_writer(self, obj: valid_types, geom_repr=ElemType.SOLID, fuse_piping=False):
        """Write current assembly to STEP file"""
        valid_geom_repr = [ElemType.SOLID, ElemType.SHELL, ElemType.LINE]
        if geom_repr not in valid_geom_repr:
            raise ValueError(f'Invalid geom_repr: "{geom_repr}". Must be in "{valid_geom_repr}"')

        if issubclass(type(obj), Shape):
            self.add_geom(obj.geom, obj)
        elif type(obj) in (Beam, Plate, Wall):
            self.export_structural(obj, geom_repr)
        elif type(obj) is Pipe:
            self.export_piping(obj, geom_repr, fuse_piping)
        elif type(obj) in (Part, Assembly):
            for sub_obj in obj.get_all_physical_objects():
                if type(sub_obj) in (Plate, Beam, Wall):
                    self.export_structural(sub_obj, geom_repr)
                elif type(sub_obj) in (Pipe,):
                    self.export_piping(sub_obj, geom_repr, fuse_piping)
                elif issubclass(type(sub_obj), Shape):
                    self.add_geom(sub_obj.geom, sub_obj)
                else:
                    raise ValueError("Unknown Geometry type")

    def write_to_file(self, destination_file, silent):
        destination_file = pathlib.Path(destination_file).with_suffix(".stp")
        os.makedirs(destination_file.parent, exist_ok=True)

        status = self.writer.Write(str(destination_file))
        if int(status) > int(IFSelect_RetError):
            raise Exception("Error during write operation")
        if silent is False:
            print(f'step file created at "{destination_file}"')

    def add_geom(self, geom, obj):
        from ada.concepts.transforms import Placement
        from ada.core.utils import vector_length

        from .utils import transform_shape

        name = obj.name if obj.name is not None else next(shp_names)
        Interface_Static_SetCVal("write.step.product.name", name)

        # Transform geometry
        res = obj.placement.absolute_placement()
        if vector_length(res - Placement().origin) > 0:
            geom = transform_shape(geom, transform=tuple(res))

        try:
            stat = self.writer.Transfer(geom, STEPControl_AsIs)
        except BaseException as e:
            logging.info(f"Passing {obj} due to {e}")
            return None

        if int(stat) > int(IFSelect_RetError):
            raise Exception("Some Error occurred")

        item = stepconstruct_FindEntity(self.fp, geom)
        if not item:
            logging.debug("STEP item not found for FindEntity")
        else:
            item.SetName(TCollection_HAsciiString(name))

    def export_structural(self, stru_obj: Union[Plate, Beam, Wall], geom_repr):
        if geom_repr == ElemType.SHELL:
            self.add_geom(stru_obj.shell, stru_obj)
        elif geom_repr == ElemType.LINE:
            self.add_geom(stru_obj.line, stru_obj)
        else:
            self.add_geom(stru_obj.solid, stru_obj)

    def export_piping(self, pipe: Pipe, geom_repr, fuse_shapes=False):
        result = None
        for pipe_seg in pipe.segments:
            if geom_repr == ElemType.LINE:
                geom = pipe_seg.line
            elif geom_repr == ElemType.SHELL:
                geom = pipe_seg.shell
            else:
                geom = pipe_seg.solid

            if fuse_shapes is True:
                if result is None:
                    result = geom
                else:
                    result = BRepAlgoAPI_Fuse(result, geom).Shape()
            else:
                self.add_geom(geom, pipe)

        if fuse_shapes is True:
            self.add_geom(result, pipe)



from typing import Dict

from ada import FEM
from ada.fem import FemSection
from ada.sections import SectionCat


def sections_str(fem: FEM):
    space = 20 * " "

    box = " BOX{id:>16}{h:>6.3f}{t_w:>10.3f}{t_ftop:>8.3f}{t_fbtn:>8.3f}{w_top:>8.3f}\n"
    tub = " PIPE{id:>14}{d:>14.3f}{wt:>14.3f}\n"
    ipe = " IHPROFIL{id:>13}{h:>13.3f}{t_w:>13.3f}{w_top:>13.3f}{t_ftop:>13.3f}{w_btn:>13.3f}{t_fbtn:>13.3f}\n"

    box_str = f"' Box Profiles\n'{space}Geom ID     H     T-sid   T-bot   T-top   Width   Sh_y Sh_z\n"
    tub_str = f"' Tubulars\n'{space}Geom ID       Do         Thick   (Shear_y   Shear_z      Diam2 )\n"
    circ_str = f"' Circulars\n'{space}Geom ID       Do         Thick   (Shear_y   Shear_z      Diam2 )\n"
    ip_str = f"' I-Girders\n'{space}Geom ID     H     T-web    W-top   T-top    W-bot   T-bot Sh_y Sh_z\n"
    tp_str = f"' T-profiles\n'{space}Geom ID     H     T-web    W-top   T-top    W-bot   T-bot Sh_y Sh_z\n"
    ang_str = f"' HP profiles\n'{space}Geom ID     H     T-web    W-top   T-top    W-bot   T-bot Sh_y Sh_z\n"
    cha_str = f"' Channels\n'{space}Geom ID     H     T-web    W-top   T-top    W-bot   T-bot Sh_y Sh_z\n"
    gens_str = f"' General Beams\n'{space}Geom ID     \n"

    sections: Dict[int, FemSection] = {fs.fem_sec.id: fs.fem_sec for fs in fem.elements.lines}
    for s_id, fs in sorted(sections.items()):
        s = fs.section
        if SectionCat.is_box_profile(s):
            # BOX      100000001    0.500   0.016   0.016   0.016    0.500
            box_str += box.format(
                id=fs.id,
                h=s.h,
                t_w=s.t_w,
                t_ftop=s.t_ftop,
                t_fbtn=s.t_fbtn,
                w_top=s.w_top,
            )
        elif SectionCat.is_tubular_profile(s):
            # PIPE      60000001       1.010       0.045
            tub_str += tub.format(id=fs.id, d=s.r * 2, wt=s.wt)
        elif SectionCat.is_circular_profile(s):
            # PIPE      60000001       1.010       0.045
            circ_str += tub.format(id=fs.id, d=s.r * 2, wt=s.r * 0.99)
        elif SectionCat.is_i_profile(s):
            # IHPROFIL     11011    0.590   0.013    0.300   0.025    0.300   0.025
            ip_str += ipe.format(
                id=fs.id,
                h=s.h,
                t_w=s.t_w,
                w_top=s.w_top,
                t_ftop=s.t_ftop,
                w_btn=s.w_btn,
                t_fbtn=s.t_fbtn,
            )
        elif SectionCat.is_t_profile(s):
            print(f'T-Profiles currently not considered. Relevant for bm id "{fs.id}". Will use IPE for now')
            # IHPROFIL     11011    0.590   0.013    0.300   0.025    0.300   0.025
            tp_str += ipe.format(
                id=fs.id,
                h=s.h,
                t_w=s.t_w,
                w_top=s.w_top,
                t_ftop=s.t_ftop,
                w_btn=s.w_btn,
                t_fbtn=s.t_fbtn,
            )

        elif SectionCat.is_angular(s):
            print(f'Angular-Profiles are not supported by USFOS. Bm "{fs.id}" will use GENBEAM')
            gens_str += general_beam(fs)
            # raise ValueError('Angular profiles currently not considered. Relevant for bm id "{}"'.format(fs.id))

        elif SectionCat.is_channel_profile(s):
            print(f'Channel-Profiles are not supported by USFOS. Bm "{fs.id}" will use GENBEAM')
            gens_str += general_beam(fs)
            # raise ValueError('Channel profiles currently not considered. Relevant for bm id "{}"'.format(fs.id))

        elif SectionCat.is_general(s):
            gens_str += general_beam(fs)
        else:
            raise ValueError(f'Unknown section string "{s.type}"')

    return box_str + ip_str + tp_str + ang_str + cha_str + tub_str + circ_str + gens_str


def general_beam(fs: FemSection) -> str:
    s = fs.section
    p = s.properties
    return (
        f" GENBEAM{fs.id:>11}{p.Ax:>11.3E}{p.Ix:>11.3E}{p.Iy:>11.3E}{p.Iz:>11.3E}\n{p.Wxmin:>11.3E}{p.Wymin:>11.3E}"
        f"{p.Wzmin:>11.3E}{p.Shary:>11.3E}\n{p.Sharz:>11.3E}\n"
    )



import re

from ada.fem.formats.utils import get_ff_regex

re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL

# Nodes
re_gnode_in = get_ff_regex("GNODE", "nodex", "nodeno", "ndof", "odof")
re_gcoord_in = get_ff_regex("GCOORD", "id", "x", "y", "z")

# Elements
re_gelmnt = get_ff_regex("GELMNT1", "elnox", "elno", "eltyp", "eltyad", "nids")
re_gelref1 = get_ff_regex(
    "GELREF1",
    "elno",
    "matno",
    "addno",
    "intno",
    "mintno",
    "strano",
    "streno",
    "strepono",
    "geono",
    "fixno",
    "eccno",
    "transno",
    "members|",
)

# Beam Sections
# GIORH (I-section)
# GUSYI (unsymm.I-section)
# GCHAN (Channel section)
# GBOX (Box section)
# GPIPE (Pipe section)
# GBARM (Massive bar)
# GTONP (T on plate)
# GDOBO (Double box)
# GLSEC (L section)
# GIORHR
# GCHANR
# GLSECR
re_sectnames = get_ff_regex("TDSECT", "nfield", "geono", "codnam", "codtxt", "set_name")
re_gbeamg = get_ff_regex(
    "GBEAMG",
    "geono",
    "comp",
    "area",
    "ix",
    "iy",
    "iz",
    "iyz",
    "wxmin",
    "wymin",
    "wzmin",
    "shary",
    "sharz",
    "shceny",
    "shcenz",
    "sy",
    "sz",
    "wy|",
    "wz|",
    "fabr|",
)
re_giorh = get_ff_regex(
    "GIORH ",
    "geono",
    "hz",
    "ty",
    "bt",
    "tt",
    "bb",
    "tb",
    "sfy",
    "sfz",
    "NLOBYT|",
    "NLOBYB|",
    "NLOBZ|",
)
re_gbox = get_ff_regex("GBOX", "geono", "hz", "ty", "tb", "tt", "by", "sfy", "sfz")
re_gpipe = get_ff_regex("GPIPE", "geono", "di", "dy", "t", "sfy", "sfz")
re_gbarm = get_ff_regex("GBARM", "geono", "hz", "bt", "bb", "sfy", "sfz")

# Coordinate System
re_lcsys = get_ff_regex("GUNIVEC", "transno", "unix", "uniy", "uniz")

# Shell section
re_thick = get_ff_regex("GELTH", "geono", "th")

# Other
re_bnbcd = get_ff_regex("BNBCD", "nodeno", "ndof", "content")
re_belfix = get_ff_regex(
    "BELFIX",
    "fixno",
    "opt",
    "trano",
    "unused",
    "a1|",
    "a2|",
    "a3|",
    "a4|",
    "a5|",
    "a6|",
)
re_mgsprng = get_ff_regex("MGSPRNG", "matno", "ndof", "bulk")
re_bnmass = get_ff_regex("BNMASS", "nodeno", "ndof", "m1", "m2", "m3", "m4", "m5", "m6")
re_mgmass = get_ff_regex("MGMASS", "matno", "ndof", "bulk")
re_geccen = get_ff_regex("GECCEN", "eccno", "ex", "ey", "ez")
re_bldep = get_ff_regex("BLDEP", "slave", "master", "nddof", "ndep", "bulk")
re_setmembs = get_ff_regex("GSETMEMB", "nfield", "isref", "index", "istype", "isorig", "members")
re_setnames = get_ff_regex("TDSETNAM", "nfield", "isref", "codnam", "codtxt", "set_name")

# Materials
re_matnames = get_ff_regex("TDMATER", "nfield", "geo_no", "codnam", "codtxt", "name")
re_misosel = get_ff_regex("MISOSEL", "matno", "young", "poiss", "rho", "damp", "alpha", "iyield", "yield")
re_morsmel = get_ff_regex(
    "MORSMEL",
    "matno",
    "q1",
    "q2",
    "q3",
    "rho",
    "d11",
    "d21",
    "d22",
    "d31",
    "d32",
    "d33",
    "ps1",
    "ps2",
    "damp1",
    "damp2",
    "alpha1",
    "alpha2",
)

# Results
re_rsumreac = get_ff_regex("RSUMREAC", "nfield", "ires", "ircomp", "x", "y", "z", "rx", "ry", "rz")
re_rvnodrea = get_ff_regex(
    "RVNODREA", "nfield", "ires", "inod", "irrea|", "irboc|", "itrans|", "F1|", "F2|", "F3|", "F4|", "F5|", "F6|"
)
re_rvnoddis = get_ff_regex(
    "RVNODDIS", "nfield", "ires", "inod", "irdva|", "itrans|", "U1|", "U2|", "U3|", "U4|", "U5|", "U6|"
)
re_rdrescmb = get_ff_regex("RDRESCMB", "nfield", "ires", "complx", "nres", "bulk")
re_rvforces = get_ff_regex("RVFORCES", "nfield", "ires", "ielno", "ispalt", "irforc|", "bulk|")
re_rdforces = get_ff_regex("RDFORCES", "nfield", "irforc", "lenrec", "bulk|")
re_tdresref = get_ff_regex("TDRESREF", "nfield", "ires", "codnam", "codtxt", "name")
re_tdload = get_ff_regex("TDLOAD", "nfield", "llc", "codnam", "codtxt", "name")



from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from typing import Iterable, List, Union

import gmsh

from ada import FEM, Beam, Pipe, Plate, Shape
from ada.base.physical_objects import BackendGeom
from ada.concepts.containers import Nodes
from ada.config import Settings
from ada.fem import Elem
from ada.fem.containers import FemElements
from ada.fem.shapes import ElemType
from ada.ifc.utils import create_guid


@dataclass
class GmshOptions:
    # Mesh
    Mesh_Algorithm: int = 6
    Mesh_ElementOrder: int = 1
    Mesh_Algorithm3D: int = 1
    Mesh_MeshSizeMin: float = None
    Mesh_MeshSizeMax: float = 0.1
    Mesh_SecondOrderIncomplete: int = 1
    Mesh_Smoothing: int = 3
    Mesh_RecombinationAlgorithm: int = None
    # Curvature Options
    Mesh_MeshSizeFromCurvature: bool = False
    Mesh_MinimumElementsPerTwoPi: int = 12
    # Geometry
    Geometry_Tolerance: float = 1e-5
    Geometry_OCCImportLabels: int = 1
    Geometry_OCCMakeSolids: int = None
    # General (UI)
    General_ColorScheme: int = 3
    General_Orthographic: int = 0  # 0 Means perspective projection
    General_Terminal: int = 1

    def get_as_dict(self):
        return {key.replace("_", "."): value for key, value in vars(self).items() if value is not None}


@dataclass
class GmshTask:
    ada_obj: List[Union[Shape, Beam, Plate]]
    geom_repr: str
    mesh_size: float
    options: GmshOptions = GmshOptions()


@dataclass
class CutPlane:
    origin: tuple
    dx: float = 10
    dy: float = 10
    plane: str = "XY"
    normal: tuple = None
    cut_objects: List[GmshData] = None
    gmsh_id: int = None


@dataclass
class GmshData:
    entities: Iterable
    geom_repr: str
    order: int
    obj: Union[Shape, Beam, Plate, Pipe]
    mesh_size: float = None


class GmshSession:
    def __init__(self, silent=False, persist=True, options: GmshOptions = GmshOptions()):
        logging.debug("init method called")
        self._gmsh = None
        self.options = options
        self.cutting_planes: List[CutPlane] = []
        self.silent = silent
        if silent is True:
            self.options.General_Terminal = 0
        self.persist = persist
        self.model_map: dict[Union[Shape, Beam, Plate, Pipe], GmshData] = dict()

    def add_obj(
        self,
        obj: Union[BackendGeom, Shape, Beam, Plate, Pipe],
        geom_repr=ElemType.SOLID,
        el_order=1,
        silent=True,
        mesh_size=None,
        build_native_lines=False,
        point_tol=Settings.point_tol,
    ):
        from ada.core.utils import Counter

        from .utils import build_bm_lines

        self.apply_settings()
        temp_dir = Settings.temp_dir
        os.makedirs(temp_dir, exist_ok=True)
        name = f"{obj.name}_{create_guid()}"

        def export_as_step(export_obj):
            export_obj.to_stp(temp_dir / name, geom_repr=geom_repr, silent=silent, fuse_piping=True)
            ents = self.model.occ.importShapes(str(temp_dir / f"{name}.stp"))
            return ents

        if issubclass(type(obj), Shape) and geom_repr != ElemType.SOLID:
            logging.info(f"geom_repr for object type {type(obj)} must be solid. Changing to that now")
            geom_repr = ElemType.SOLID

        if build_native_lines is True and geom_repr == ElemType.LINE and type(obj) is Beam:
            # midpoints = obj.calc_con_points()
            entities = build_bm_lines(self.model, obj, point_tol)
            # if len(midpoints) > 0:
            #
            # else:
            #     entities = export_as_step(obj)
        else:
            entities = export_as_step(obj)
        #
        # self.model.geo.synchronize()
        # self.model.occ.synchronize()
        obj_name = Counter(1, f"{obj.name}_")
        for dim, ent in entities:
            ent_name = next(obj_name)
            self.model.set_physical_name(dim, ent, ent_name)
            self.model.set_entity_name(dim, ent, ent_name)
        self.model.occ.synchronize()
        self.model.geo.synchronize()

        gmsh_data = GmshData(entities, geom_repr, el_order, obj, mesh_size=mesh_size)
        self.model_map[obj] = gmsh_data
        return gmsh_data

    def add_cutting_plane(self, cut_plane: CutPlane, cut_objects: List[GmshData] = None):
        if cut_objects is not None:
            if cut_plane.cut_objects is None:
                cut_plane.cut_objects = []
            cut_plane.cut_objects += cut_objects
        self.cutting_planes.append(cut_plane)

    def make_cuts(self):
        geom_repr_map = {ElemType.SOLID: 3, ElemType.SHELL: 2, ElemType.LINE: 1}

        for cut in self.cutting_planes:
            x, y, z = cut.origin
            rect = self.model.occ.addRectangle(x, y, z, cut.dx, cut.dy)
            cut.gmsh_id = rect
            for obj in cut.cut_objects:
                res, _ = self.model.occ.fragment(obj.entities, [(2, rect)], removeTool=True)
                cut_geom_dim = geom_repr_map[obj.geom_repr]
                replaced_entities = [(dim, r) for dim, r in res if r != rect and dim == cut_geom_dim]
                obj.entities = replaced_entities
            self.model.occ.remove([(2, rect)], True)

        # rem_ids = [(2, c.gmsh_id) for c in self.cutting_planes]
        # self.model.occ.remove(rem_ids, True)
        self.model.occ.synchronize()

    def split_plates_by_beams(self):
        from ada.core.clash_check import (
            filter_away_beams_along_plate_edges,
            find_beams_connected_to_plate,
        )

        beams = [obj for obj in self.model_map.keys() if type(obj) is Beam]
        plates = [obj for obj in self.model_map.keys() if type(obj) is Plate]
        for pl in plates:
            pl_gmsh_obj = self.model_map[pl]
            for pl_dim, pl_ent in pl_gmsh_obj.entities:
                intersecting_beams = []
                int_bm_map = dict()
                all_contained_beams = find_beams_connected_to_plate(pl, beams)
                inside_beams = filter_away_beams_along_plate_edges(pl, all_contained_beams)
                for bm in inside_beams:
                    bm_gmsh_obj = self.model_map[bm]
                    for li_dim, li_ent in bm_gmsh_obj.entities:
                        intersecting_beams.append((li_dim, li_ent))
                        int_bm_map[(li_dim, li_ent)] = bm_gmsh_obj
                # Using Embed fails during meshing

                # res = self.model.mesh.embed(1, [t for e,t in intersecting_beams], 2, pl_ent)

                res, res_map = self.model.occ.fragment(intersecting_beams, [(pl_dim, pl_ent)])
                replaced_pl_entities = [(dim, r) for dim, r in res if dim == 2]
                for i, int_bm in enumerate(intersecting_beams):
                    bm_gmsh_obj = int_bm_map[int_bm]
                    new_ents = res_map[i]
                    bm_gmsh_obj.entities = new_ents
                pl_gmsh_obj.entities = replaced_pl_entities

                self.model.occ.synchronize()

    def mesh(self, size: float = None):
        if self.silent is True:
            self.options.General_Terminal = 0

        if size is not None:
            self.options.Mesh_MeshSizeMax = size
            if self.options.Mesh_MeshSizeFromCurvature is False:
                self.options.Mesh_MeshSizeMin = size

        self.apply_settings()
        self.model.geo.synchronize()
        self.model.mesh.setRecombine(3, -1)
        self.model.mesh.generate(3)
        self.model.mesh.removeDuplicateNodes()

    def get_fem(self) -> FEM:
        from .utils import (
            add_fem_sections,
            get_elements_from_entities,
            get_nodes_from_gmsh,
        )

        fem = FEM("AdaFEM")
        gmsh_nodes = get_nodes_from_gmsh(self.model, fem)
        fem.nodes = Nodes(gmsh_nodes, parent=fem)

        def add_obj_to_elem_ref(el: Elem, obj: Union[Shape, Beam, Plate, Pipe]):
            el.refs.append(obj)

        # Get Elements
        elements = []
        for gmsh_data in self.model_map.values():
            entity_elements = get_elements_from_entities(self.model, gmsh_data.entities, fem)
            gmsh_data.obj.elem_refs = entity_elements
            [add_obj_to_elem_ref(el, gmsh_data.obj) for el in entity_elements]
            elements += entity_elements
        fem.elements = FemElements(elements, fem_obj=fem)

        # Add FEM sections
        for model_obj, gmsh_data in self.model_map.items():
            add_fem_sections(self.model, fem, model_obj, gmsh_data)

        fem.nodes.renumber()
        fem.elements.renumber()
        return fem

    def apply_settings(self):
        if self.options is not None:
            for setting, value in self.options.get_as_dict().items():
                self.gmsh.option.setNumber(setting, value)
        self.model.geo.synchronize()

    def open_gui(self):
        self.gmsh.fltk.run()

    def __enter__(self):
        logging.debug("Starting GMSH session")

        self._gmsh = gmsh
        self.gmsh.initialize()
        # self.model.add("ada")
        self.apply_settings()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        logging.debug("Closing GMSH")
        self.gmsh.finalize()

    @property
    def gmsh(self) -> gmsh:
        return self._gmsh

    @property
    def model(self) -> "gmsh.model":
        return self.gmsh.model



from __future__ import annotations

from .common import Amplitude, Csys
from .constraints import Bc, Constraint, PredefinedField
from .elements import Connector, Elem, Mass, Spring
from .interactions import Interaction, InteractionProperty
from .loads import Load, LoadCase, LoadPressure
from .outputs import FieldOutput, HistOutput
from .sections import ConnectorSection, FemSection
from .sets import FemSet
from .steps import StepEigen, StepExplicit, StepImplicit, StepSteadyState
from .surfaces import Surface

__all__ = [
    "Amplitude",
    "Bc",
    "Csys",
    "InteractionProperty",
    "Interaction",
    "StepSteadyState",
    "StepEigen",
    "StepImplicit",
    "StepExplicit",
    "Surface",
    "Elem",
    "Connector",
    "Constraint",
    "PredefinedField",
    "FemSet",
    "Mass",
    "HistOutput",
    "FieldOutput",
    "ConnectorSection",
    "Load",
    "LoadPressure",
    "LoadCase",
    "FemSection",
    "Spring",
]



from __future__ import annotations

from typing import Union

from .common import Csys, FemBase
from .sets import FemSet
from .surfaces import Surface


class BcTypes:
    DISPL = "displacement"
    VELOCITY = "velocity"
    CONN_DISPL = "connector displacement"
    CONN_VEL = "connector velocity"
    ENCASTRE = "symmetry/antisymmetry/encastre"
    DISPL_ROT = "displacement/rotation"
    VELOCITY_ANGULAR = "velocity/angular velocity"

    all = [DISPL, VELOCITY, CONN_DISPL, CONN_VEL, ENCASTRE, DISPL_ROT, VELOCITY_ANGULAR]


class PreDefTypes:
    VELOCITY = "VELOCITY"
    INITIAL_STATE = "INITIAL STATE"

    all = [VELOCITY, INITIAL_STATE]


class ConstraintTypes:
    COUPLING = "coupling"
    TIE = "tie"
    RIGID_BODY = "rigid body"
    MPC = "mpc"
    SHELL2SOLID = "shell2solid"


class Bc(FemBase):
    TYPES = BcTypes

    def __init__(
        self,
        name,
        fem_set: FemSet,
        dofs,
        magnitudes=None,
        bc_type=BcTypes.DISPL,
        amplitude_name=None,
        init_condition=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._fem_set = fem_set
        self._dofs = dofs if type(dofs) is list else [dofs]
        if magnitudes is None:
            self._magnitudes = [None] * len(self._dofs)
        else:
            self._magnitudes = magnitudes if type(magnitudes) is list else [magnitudes]
        self.type = bc_type.lower()
        self._amplitude_name = amplitude_name
        self._init_condition = init_condition

    def add_init_condition(self, init_condition):
        self._init_condition = init_condition

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.lower() not in BcTypes.all:
            raise ValueError(f'BC type "{value}" is not yet supported')
        self._type = value.lower()

    @property
    def fem_set(self) -> FemSet:
        return self._fem_set

    @property
    def dofs(self):
        return self._dofs

    @property
    def magnitudes(self):
        return self._magnitudes

    @property
    def amplitude_name(self):
        return self._amplitude_name

    def __repr__(self):
        return f'Bc("{self.name}", type="{self.type}", dofs={self.dofs}, fem_set="{self.fem_set.name}")'


class Constraint(FemBase):
    TYPES = ConstraintTypes

    def __init__(
        self,
        name,
        con_type,
        m_set: FemSet,
        s_set: Union[FemSet, Surface],
        dofs=None,
        pos_tol=None,
        mpc_type=None,
        csys: Csys = None,
        parent=None,
        metadata=None,
        influence_distance: float = None,
    ):
        super().__init__(name, metadata, parent)
        self._con_type = con_type
        self._m_set = m_set
        self._s_set = s_set
        self._dofs = [1, 2, 3, 4, 5, 6] if dofs is None else dofs
        self._pos_tol = pos_tol
        self._mpc_type = mpc_type
        self._csys = csys
        self._influence_distance = influence_distance

    @property
    def type(self):
        return self._con_type

    @property
    def m_set(self) -> FemSet:
        return self._m_set

    @property
    def s_set(self) -> FemSet:
        return self._s_set

    @property
    def dofs(self):
        return self._dofs

    @property
    def pos_tol(self):
        return self._pos_tol

    @property
    def csys(self) -> Csys:
        return self._csys

    @property
    def mpc_type(self):
        return self._mpc_type

    @property
    def influence_distance(self):
        return self._influence_distance

    def __repr__(self):
        return f'Constraint("{self.type}", m: "{self.m_set.name}", s: "{self.s_set.name}", dofs: "{self.dofs}")'


class PredefinedField(FemBase):
    TYPES = PreDefTypes

    def __init__(
        self,
        name,
        field_type,
        fem_set=None,
        dofs=None,
        magnitude=None,
        initial_state_file=None,
        initial_state_part=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self.type = field_type
        self._fem_set = fem_set
        self._dofs = dofs
        self._magnitude = magnitude
        self._initial_state_part = initial_state_part
        self._initial_state_file = initial_state_file
        if self.initial_state_file is not None:
            self.initial_state_part.fem.initial_state = self

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.upper() not in PreDefTypes.all:
            raise ValueError(f'The field type "{value.upper()}" is currently not supported')
        self._type = value.upper()

    @property
    def fem_set(self) -> FemSet:
        """

        :return:
        :rtype: FemSet
        """
        return self._fem_set

    @property
    def dofs(self):
        return self._dofs

    @property
    def magnitude(self):
        return self._magnitude

    @property
    def initial_state_part(self):
        """

        :rtype: ada.Part
        """
        return self._initial_state_part

    @property
    def initial_state_file(self):
        return self._initial_state_file



import logging

from ada import FEM
from ada.fem import Constraint
from ada.fem.common import LinDep

from .write_utils import write_ff


def constraint_str(fem: FEM) -> str:
    out_str = ""
    for constraint in fem.constraints:
        if constraint.type == constraint.TYPES.COUPLING:
            out_str += write_coupling(constraint)
        else:
            raise logging.error(f'Constraint type "{constraint.type}" is not yet supported')

    return out_str


def write_coupling(constraint: Constraint) -> str:
    out_str = ""
    master = constraint.m_set.members[0]
    master_id = master.id
    for node in constraint.s_set.members:
        slave_id = node.id
        res = LinDep(master.p, node.p)
        lin_deps = [(slave_id, master_id, 3, 9)]
        for lin_dep_rel in res.to_integer_list():
            lin_deps.append(tuple(list(lin_dep_rel) + [0.0]))
        out_str += write_ff("BLDEP", lin_deps)

    return out_str



import logging
import os
import pathlib
from typing import List, Union

from ada.fem import StepEigen
from ada.fem.concepts.eigenvalue import EigenDataSummary, EigenMode
from ada.fem.formats.utils import DatFormatReader
from ada.fem.results import Results


def get_eigen_data(dat_file: Union[str, os.PathLike]) -> EigenDataSummary:
    dtr = DatFormatReader()

    re_compiled = dtr.compile_ff_re([int] + [float] * 3, separator=";")

    eig_str = "printofeigenvalues"
    eig_res = dtr.read_data_lines(dat_file, re_compiled, eig_str, split_data=True)
    eigen_modes: List[EigenMode] = []

    # Note! participation factors and effective modal mass are each deconstructed into 6 degrees of freedom
    for mode, eig_value, eig_freq, period in eig_res:
        eig_output = dict(
            eigenvalue=eig_value.replace(";", ""),
            f_hz=eig_freq.replace(";", ""),
        )
        eigen_modes.append(EigenMode(no=mode.replace(";", ""), **eig_output))

    return EigenDataSummary(eigen_modes)


def read_sesam_results(results: Results, file_ref: pathlib.Path, overwrite):
    dat_file = (file_ref.parent / "SESTRA").with_suffix(".LIS")
    if dat_file.exists() and type(results.assembly.fem.steps[0]) == StepEigen:
        results.eigen_mode_data = get_eigen_data(dat_file)

    logging.error("Result mesh data extraction is not supported for sesam")
    return None



import logging
import pathlib

import h5py
import meshio
import numpy as np

from ada.fem import StepEigen
from ada.fem.concepts.eigenvalue import EigenDataSummary, EigenMode
from ada.fem.elements import ElemShape
from ada.fem.results import Results

from .reader import med_to_fem


def get_eigen_data(rmed_file) -> EigenDataSummary:
    f = h5py.File(rmed_file)
    modes = f.get("CHA/modes___DEPL")
    eigen_modes = []

    for mname, m in modes.items():
        mode = m.attrs["NDT"]
        freq = m.attrs["PDT"]
        eigen_modes.append(EigenMode(mode, f_hz=freq))

    return EigenDataSummary(eigen_modes)


def get_eigen_frequency_deformed_meshes(rmed_file):
    fem = med_to_fem(rmed_file, "temp")
    f = h5py.File(rmed_file)
    modes = f.get("CHA/modes___DEPL")
    nodes = fem.nodes.to_np_array()
    eig_deformed_meshes = []

    for mname, m in modes.items():
        res = m["NOE"]["MED_NO_PROFILE_INTERNAL"]["CO"][()]
        dofs = res.reshape(len(fem.nodes), 6)
        eig_deformed_meshes.append(nodes + np.delete(dofs, np.s_[2:5], 1))
        mode = m.attrs["NDT"]
        freq = m.attrs["PDT"]
        print(mode, freq)

    # TODO: Figure out what kind of information is needed for animating frames in threejs/blender
    return fem, eig_deformed_meshes


def read_code_aster_results(results: Results, file_ref: pathlib.Path, overwrite):
    if type(results.assembly.fem.steps[0]) is StepEigen:
        results.eigen_mode_data = get_eigen_data(file_ref)
    fem = med_to_fem(file_ref, "temp")
    if any([x.type in ElemShape.TYPES.tri7 for x in fem.elements.shell]):
        logging.error("Meshio does not support 7 node Triangle elements yet")
        return None
    return meshio.read(file_ref, "med")



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Tuple

from ada.base.non_phyical_objects import Backend
from ada.concepts.curves import CurvePoly
from ada.config import Settings

from .categories import BaseTypes, SectionCat


class Section(Backend):
    TYPES = BaseTypes

    def __init__(
        self,
        name,
        sec_type=None,
        h=None,
        w_top=None,
        w_btn=None,
        t_w=None,
        t_ftop=None,
        t_fbtn=None,
        r=None,
        wt=None,
        sec_id=None,
        parent=None,
        sec_str=None,
        from_str=None,
        outer_poly=None,
        inner_poly=None,
        genprops=None,
        metadata=None,
        units="m",
        ifc_elem=None,
        guid=None,
        refs=None,
    ):
        super(Section, self).__init__(name, guid, metadata, units, ifc_elem=ifc_elem, parent=parent)
        self._type = sec_type
        self._h = h
        self._w_top = w_top
        self._w_btn = w_btn
        self._t_w = t_w
        self._t_ftop = t_ftop
        self._t_fbtn = t_fbtn
        self._r = r
        self._wt = wt
        self._id = sec_id
        self._outer_poly = outer_poly
        self._inner_poly = inner_poly
        self._sec_str = sec_str

        self._ifc_profile = None
        self._ifc_beam_type = None

        if ifc_elem is not None:
            props = self._import_from_ifc_profile(ifc_elem)
            self.__dict__.update(props.__dict__)

        if from_str is not None:
            from ada.sections.utils import interpret_section_str

            if units == "m":
                scalef = 0.001
            elif units == "mm":
                scalef = 1.0
            else:
                raise ValueError(f'Unknown units "{units}"')
            sec, tap = interpret_section_str(from_str, scalef, units=units)
            self.__dict__.update(sec.__dict__)
        elif outer_poly:
            self._type = "poly"

        self._genprops = None
        self._refs = refs if refs is not None else []
        if genprops is not None:
            genprops.parent = self
            self._genprops = genprops

    def __eq__(self, other):
        for key, val in self.__dict__.items():
            if "parent" in key or "_ifc" in key or key in ["_sec_id", "_guid"]:
                continue
            oval = other.__dict__[key]
            if oval != val:
                return False

        return True

    def _generate_ifc_section_data(self):
        from ada.ifc.export_beam_sections import export_beam_section

        return export_beam_section(self)

    def _import_from_ifc_profile(self, ifc_elem):
        from ada.ifc.import_beam_section import import_section_from_ifc

        self._ifc_profile = ifc_elem
        return import_section_from_ifc(ifc_elem)

    @property
    def type(self):
        return self._type

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if type(value) is not int:
            raise ValueError
        self._id = value

    @property
    def h(self):
        return self._h

    @property
    def w_top(self):
        return self._w_top

    @w_top.setter
    def w_top(self, value):
        """Width of top flange"""
        self._w_top = value

    @property
    def w_btn(self):
        """Width of bottom flange"""
        return self._w_btn

    @w_btn.setter
    def w_btn(self, value):
        self._w_btn = value

    @property
    def t_w(self):
        """Thickness of web"""
        return self._t_w

    @property
    def t_ftop(self):
        """Thickness of top flange"""
        return self._t_ftop

    @property
    def t_fbtn(self):
        """Thickness of bottom flange"""
        return self._t_fbtn

    @property
    def r(self) -> float:
        """Radius (Outer)"""
        return self._r

    @property
    def wt(self) -> float:
        """Wall thickness"""
        return self._wt

    @property
    def sec_str(self):
        def s(x):
            return x / 0.001

        if self.type in SectionCat.box + SectionCat.igirders + SectionCat.tprofiles + SectionCat.shs + SectionCat.rhs:
            sec_str = "{}{:g}x{:g}x{:g}x{:g}".format(self.type, s(self.h), s(self.w_top), s(self.t_w), s(self.t_ftop))
        elif self.type in SectionCat.tubular:
            sec_str = "{}{:g}x{:g}".format(self.type, s(self.r), s(self.wt))
        elif self.type in SectionCat.circular:
            sec_str = "{}{:g}".format(self.type, s(self.r))
        elif self.type in SectionCat.angular:
            sec_str = "{}{:g}x{:g}".format(self.type, s(self.h), s(self.t_w))
        elif self.type in SectionCat.iprofiles:
            sec_str = self._sec_str
        elif self.type in SectionCat.channels:
            sec_str = "{}{:g}".format(self.type, s(self.h))
        elif self.type in SectionCat.general:
            sec_str = "{}{}".format(self.type, self.id)
        elif self.type in SectionCat.flatbar:
            sec_str = f"{self.type}{s(self.h)}x{s(self.w_top)}"
        elif self.type == "poly":
            sec_str = "PolyCurve"
        else:
            raise ValueError(f'Section type "{self.type}" has not been given a section str')
        return sec_str.replace(".", "_") if sec_str is not None else None

    @property
    def properties(self) -> GeneralProperties:
        if self._genprops is None:
            from .properties import calculate_general_properties

            self._genprops = calculate_general_properties(self)

        return self._genprops

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if self._units != value:
            from ada.core.utils import unit_length_conversion

            scale_factor = unit_length_conversion(self._units, value)

            if self.poly_inner is not None:
                self.poly_inner.scale(scale_factor, Settings.point_tol)

            if self.poly_outer is not None:
                self.poly_outer.scale(scale_factor, Settings.point_tol)

            vals = ["h", "w_top", "w_btn", "t_w", "t_ftop", "t_fbtn", "r", "wt"]

            for key in self.__dict__.keys():
                if self.__dict__[key] is not None:
                    if key[1:] in vals:
                        self.__dict__[key] *= scale_factor
            self._units = value

    @property
    def ifc_profile(self):
        if self._ifc_profile is None:
            self._ifc_profile, self._ifc_beam_type = self._generate_ifc_section_data()
        return self._ifc_profile

    @property
    def ifc_beam_type(self):
        if self._ifc_beam_type is None:
            self._ifc_profile, self._ifc_beam_type = self._generate_ifc_section_data()
        return self._ifc_beam_type

    @property
    def poly_outer(self) -> CurvePoly:
        return self._outer_poly

    @property
    def poly_inner(self) -> CurvePoly:
        return self._inner_poly

    def get_section_profile(self, is_solid=True) -> SectionProfile:
        return build_section_profile(self, is_solid)

    def _repr_html_(self):
        from IPython.display import display
        from ipywidgets import HBox

        from ada.visualize.renderer_pythreejs import SectionRenderer

        sec_render = SectionRenderer()
        fig, html = sec_render.build_display(self)
        display(HBox([fig, html]))

    @property
    def refs(self):
        """:rtype: List[ada.Beam | ada.fem.FemSection]"""
        return self._refs

    def __hash__(self):
        return hash(self.guid)

    def __repr__(self):
        if self.type in SectionCat.circular + SectionCat.tubular:
            return f"Section({self.name}, {self.type}, r: {self.r}, wt: {self.wt})"
        elif self.type in SectionCat.general:
            p = self.properties
            return f"Section({self.name}, {self.type}, Ax: {p.Ax}, Ix: {p.Ix}, Iy: {p.Iy}, Iz: {p.Iz}, Iyz: {p.Iyz})"
        else:
            return (
                f"Section({self.name}, {self.type}, h: {self.h}, w_btn: {self.w_btn}, "
                f"w_top: {self.w_top}, t_fbtn: {self.t_fbtn}, t_ftop: {self.t_ftop}, t_w: {self.t_w})"
            )


class SectionParts:
    WEB = "web"
    TOP_FLANGE = "top_fl"
    BTN_FLANGE = "btn_fl"


@dataclass
class GeneralProperties:
    parent: Section = None
    Ax: float = None
    Ix: float = None
    Iy: float = None
    Iz: float = None
    Iyz: float = None
    Wxmin: float = None
    Wymin: float = None
    Wzmin: float = None
    Shary: float = None
    Sharz: float = None
    Shceny: float = None
    Shcenz: float = None
    Sy: float = None
    Sz: float = None
    Sfy: float = 1
    Sfz: float = 1
    Cy: float = None
    Cz: float = None

    def __eq__(self, other):
        for key, val in self.__dict__.items():
            if "parent" in key:
                continue
            if other.__dict__[key] != val:
                return False

        return True


@dataclass
class SectionProfile:
    sec: Section
    is_solid: bool
    outer_curve: CurvePoly = None
    inner_curve: CurvePoly = None
    outer_curve_disconnected: List[CurvePoly] = None
    inner_curve_disconnected: List[CurvePoly] = None
    disconnected: bool = None
    shell_thickness_map: List[Tuple[str, float]] = None


def build_section_profile(sec: Section, is_solid) -> SectionProfile:
    import ada.sections.profiles as profile_builder

    section_shape_type = SectionCat.get_shape_type(sec)

    if section_shape_type in SectionCat.tubular + SectionCat.circular + SectionCat.general:
        logging.info("Tubular profiles do not need curve representations")
        return SectionProfile(sec, is_solid)

    sec_type = SectionCat.BASETYPES
    build_map = {
        sec_type.ANGULAR: profile_builder.angular,
        sec_type.IPROFILE: profile_builder.iprofiles,
        sec_type.BOX: profile_builder.box,
        sec_type.FLATBAR: profile_builder.flatbar,
        sec_type.CHANNEL: profile_builder.channel,
    }

    section_builder = build_map.get(section_shape_type, None)

    if section_builder is None and sec.poly_outer is None:
        raise ValueError("Currently geometry build is unsupported for profile type {ptype}".format(ptype=sec.type))

    if section_builder is not None:
        section_profile = section_builder(sec, is_solid)
    else:
        section_profile = SectionProfile(sec, outer_curve=sec.poly_outer, is_solid=is_solid, disconnected=False)

    return section_profile



from dataclasses import dataclass
from typing import List, Union

from ada.concepts.points import Node

from .common import FemBase
from .elements import Elem
from .sets import FemSet


class SurfTypes:
    ELEMENT = "ELEMENT"
    NODE = "NODE"

    all = [ELEMENT, NODE]


@dataclass
class ElemSurface:
    fem_set: FemSet
    side_index: int


class Surface(FemBase):
    """
    Documentation

        https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-surface.htm#simakey-r-surface__simakey-r-surface-s-datadesc5


    Parameters.

    :param name: Unique name of surface
    :param surf_type: Type of surface
    :param fem_set:
    :param weight_factor:
    :param id_refs: Explicitly defined by list of tuple [(elid/nid,spos), ..]
    :param parent:
    :param metadata:
    """

    TYPES = SurfTypes

    def __init__(
        self,
        name,
        surf_type,
        fem_set: Union[FemSet, List[FemSet]],
        weight_factor=None,
        el_face_index: Union[int, List[int]] = None,
        id_refs=None,
        parent=None,
        metadata=None,
    ):
        """:type parent: ada.FEM"""
        super().__init__(name, metadata, parent)

        self._type = surf_type.upper()

        if self.type not in SurfTypes.all:
            raise ValueError(f'Surface type "{self.type}" is currently not supported\\implemented. Valid types are')

        self._fem_set = fem_set
        if type(fem_set) is list:
            if not type(el_face_index) is list:
                raise ValueError("You cannot define a list of FemSets and not also include a List of el_face_indices")

        self._weight_factor = weight_factor
        self._el_face_index = el_face_index
        self._id_refs = id_refs

    @property
    def type(self):
        return self._type

    @property
    def fem_set(self) -> Union[FemSet, List[FemSet]]:
        return self._fem_set

    @property
    def weight_factor(self):
        return self._weight_factor

    @property
    def el_face_index(self) -> Union[int, List[int]]:
        return self._el_face_index

    @property
    def id_refs(self):
        return self._id_refs


def create_surface_from_nodes(surface_name: str, nodes: List[Node], fem):
    """:type fem: ada.FEM"""
    elements = []
    face_seq_indices = {}
    for n in nodes:
        for el in n.refs:
            if el.id == 4341:
                print("sd")
            paralell_face_index = elem_has_parallel_face(el, nodes)
            if paralell_face_index is None:
                continue

            if el not in elements:
                face_seq_indices[el] = paralell_face_index
                elements.append(el)

    fsets = []
    fset_el_face_indices = []
    for el, el_face_index in face_seq_indices.items():
        side_name = f"S{el_face_index + 1}"
        fs = FemSet(f"_{surface_name}_{el.id}_{side_name}", [el])
        if fs.name in fem.sets.elements.keys():
            fs_elem_1 = fem.sets.elements[fs.name]
            fs_elem_1.add_members([el])
        else:
            fs_elem_1 = fem.add_set(fs)
        fsets.append(fs_elem_1)
        fset_el_face_indices.append(el_face_index)

    return Surface(surface_name, Surface.TYPES.ELEMENT, fsets, el_face_index=fset_el_face_indices)


def elem_has_parallel_face(el: Elem, nodes: List[Node]):
    for i, nid_refs in enumerate(el.shape.faces_seq):
        all_face_nodes_in_plane = True
        for nid in nid_refs:
            no = el.nodes[nid]
            if no not in nodes:
                all_face_nodes_in_plane = False
                break
        if all_face_nodes_in_plane is True:
            return i
    return None



import numpy as np


def write_ff(flag: str, data):
    """
    flag = NCOD
    data = [(int, float, int, float), (float, int)]

    ->> NCOD    INT     FLOAT       INT     FLOAT
                FLOAT   INT

    :param flag:
    :param data:
    :return:
    """

    out_str = f"{flag:<8}"
    for row in data:
        v = [format_data(x) for x in row]
        if row == data[-1]:
            out_str += "".join(v) + "\n"
        else:
            out_str += "".join(v) + "\n" + 8 * " "
    return out_str


def format_data(d):
    if type(d) in (np.float64, float, int, np.uint64, np.int32) and d >= 0:
        d = make_zero(d)
        return f"  {d:<14.8E}"
    elif type(d) in (np.float64, float, int, np.uint64, np.int32) and d < 0:
        d = make_zero(d)
        return f" {d:<15.8E}"
    elif type(d) is str:
        return d
    else:
        raise ValueError(f"Unknown input {type(d)} {d}")


def make_zero(d):
    return d if abs(d) != 0.0 else 0.0



from dataclasses import dataclass

import numpy as np

from ada.config import Settings


class FemBase:
    def __init__(self, name, metadata, parent):
        """:type parent: ada.FEM"""
        self.name = name
        self.parent = parent
        self._metadata = metadata if metadata is not None else dict()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        from ada.core.utils import make_name_fem_ready

        if str.isnumeric(value[0]):
            raise ValueError("Name cannot start with numeric")

        if Settings.convert_bad_names_for_fem:
            self._name = make_name_fem_ready(value)
        else:
            self._name = value.strip()

    @property
    def parent(self):
        """:rtype: ada.FEM"""
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def metadata(self):
        return self._metadata


class CsysSystems:
    RECTANGULAR = "RECTANGULAR"
    # , 'CYLINDRICAL', 'SPHERICAL', 'Z RECTANGULAR', 'USER']


class CsysDefs:
    COORDINATES = "COORDINATES"
    NODES = "NODES"
    # ,'OFFSET TO NODES'


class Csys(FemBase):
    TYPES_SYSTEM = CsysSystems
    TYPES_DEFINITIONS = CsysDefs

    def __init__(
        self,
        name,
        definition=TYPES_DEFINITIONS.COORDINATES,
        system=TYPES_SYSTEM.RECTANGULAR,
        nodes=None,
        coords=None,
        metadata=None,
        parent=None,
    ):
        """:type parent: ada.FEM"""
        super().__init__(name, metadata, parent)
        self._definition = definition
        self._system = system
        self._nodes = nodes
        self._coords = coords

    @property
    def definition(self):
        return self._definition

    @property
    def system(self):
        return self._system

    @property
    def nodes(self):
        return self._nodes

    @property
    def coords(self):
        return self._coords

    def __repr__(self):
        content_map = dict(COORDINATES=self.coords, NODES=self.nodes)
        return f'Csys("{self.name}", "{self.definition}", {content_map[self.definition]})'


class Amplitude(FemBase):
    def __init__(self, name, x, y, smooth=None, metadata=None, parent=None):
        """:type parent: ada.FEM"""
        super().__init__(name, metadata, parent)
        self._x = x
        self._y = y
        self._smooth = smooth

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def smooth(self):
        return self._smooth


@dataclass
class LinDep:
    master: np.ndarray
    slave: np.ndarray

    Xdict: dict = None
    Ydict: dict = None
    Zdict: dict = None

    def __post_init__(self):
        diff = self.slave - self.master
        dx, dy, dz = diff

        self.X = dict(
            x=1,
            xR_z=-dy,
            xR_y=dz,
        )
        self.Y = dict(
            y=1,
            yR_z=dx,
            yR_x=-dz,
        )
        self.Z = dict(
            z=1,
            zR_x=dy,
            zR_y=-dx,
        )

    def to_integer_list(self):
        return [
            (1, 1, self.X["x"]),
            (1, 5, self.X["xR_y"]),
            (1, 6, self.X["xR_z"]),
            (2, 2, self.Y["y"]),
            (2, 4, self.Y["yR_x"]),
            (2, 6, self.Y["yR_z"]),
            (3, 3, self.Z["z"]),
            (3, 4, self.Z["zR_x"]),
            (3, 5, self.Z["zR_y"]),
        ]



import getpass
import os
import pathlib
from dataclasses import dataclass
from datetime import datetime


def _get_platform_home():
    """Home location for each platform"""
    # _platform_home = dict(win32="C:/ADA", linux="/home/ADA", linux2="/home/ADA", macos="/home/ADA")
    # return _platform_home[sys.platform]

    return pathlib.Path.home() / "ADA"


class Settings:
    """The Properties object contains all general purpose properties relevant for Parts and Assemblies"""

    point_tol = 1e-4
    precision = 6
    mtol = 1e-3
    mmtol = 1
    valid_units = ["m", "mm"]

    convert_bad_names = False
    convert_bad_names_for_fem = True
    use_occ_bounding_box_algo = False
    make_param_elbows = False
    use_param_profiles = True
    silence_display = False
    use_experimental_cache = False

    # IFC export settings
    include_ecc = True

    # FEM analysis settings
    if os.getenv("ADA_execute_dir", None) is not None:
        execute_dir = pathlib.Path(os.getenv("ADA_execute_dir", None))
    else:
        execute_dir = None

    # Visualization Settings
    use_new_visualize_api = False

    # Code Aster conversion specific settings
    ca_experimental_id_numbering = False

    debug = False
    _home = _get_platform_home()
    scratch_dir = pathlib.Path(os.getenv("ADA_scratch_dir", f"{_home}/Analyses"))
    temp_dir = pathlib.Path(os.getenv("ADA_temp_dir", f"{_home}/temp"))
    debug_dir = pathlib.Path(os.getenv("ADA_log_dir", f"{_home}/logs"))
    test_dir = pathlib.Path(os.getenv("ADA_test_dir", f"{_home}/tests"))
    tools_dir = pathlib.Path(os.getenv("ADA_tools_dir", f"{_home}/tools"))

    fem_exe_paths = dict(abaqus=None, ccx=None, sestra=None, usfos=None, code_aster=None)

    @classmethod
    def default_ifc_settings(cls):
        from ada.ifc.utils import default_settings

        return default_settings()


@dataclass
class User:
    user_id: str = getpass.getuser()
    given_name: str = None
    family_name: str = None
    middle_names: str = None
    prefix_titles: str = None
    suffix_titles: str = None
    org_id: str = "ADA"
    org_name: str = "Assembly For Design and Analysis"
    org_description: str = None
    role: str = "Engineer"
    parent = None

    def _build_ifc_elem(self):

        from ada.ifc.utils import get_org, get_person

        f = self.parent.ifc_file
        actor = f.create_entity("IfcActorRole", self.role.upper(), None, None)
        user_props = dict(
            Identification=self.user_id,
            FamilyName=self.family_name,
            GivenName=self.given_name,
            MiddleNames=self.middle_names,
            PrefixTitles=self.prefix_titles,
            SuffixTitles=self.suffix_titles,
        )
        person = get_person(f, self.user_id)
        if person is None:
            person = f.create_entity("IfcPerson", **user_props, Roles=(actor,))
        organization = get_org(f, self.org_id)
        if organization is None:
            organization = f.create_entity(
                "IfcOrganization",
                Identification=self.org_id,
                Name=self.org_name,
                Description=self.org_description,
            )
        p_o = f.create_entity("IfcPersonAndOrganization", person, organization)
        application = f.create_entity("IfcApplication", organization, "XXX", "ADA", "ADA")
        timestamp = int(datetime.now().timestamp())

        return f.create_entity("IfcOwnerHistory", p_o, application, "READWRITE", None, None, None, None, timestamp)

    def to_ifc(self):
        # Important! Needs to create unique owner_history for each use. Will cause seg fault when adding non-unique
        return self._build_ifc_elem()



from ada import Assembly
from ada.fem.exceptions.element_support import IncompatibleElements
from ada.fem.formats.utils import get_fem_model_from_assembly
from ada.fem.steps import Step


def check_compatibility(assembly: Assembly):
    from .writer import must_be_converted_to_general_section

    p = get_fem_model_from_assembly(assembly)
    step = assembly.fem.steps[0] if len(assembly.fem.steps) > 0 else None

    if step is not None:
        if assembly.fem.steps[0].type == Step.TYPES.EIGEN:
            for line in p.fem.elements.lines:
                if must_be_converted_to_general_section(line.fem_sec.section.type):
                    raise IncompatibleElements("Calculix does not support general beam elements in Eigenvalue analysis")



import logging
import os

import h5py
import numpy as np

from ada.concepts.containers import Nodes
from ada.concepts.levels import FEM, Assembly, Part
from ada.fem import Elem, FemSet
from ada.fem.containers import FemElements, FemSets

from .common import med_to_ada_type


def read_fem(assembly: Assembly, fem_file: os.PathLike, fem_name: str = None):
    fem = med_to_fem(fem_file, fem_name)
    if fem_name is None:
        fem_name = fem.name
    assembly.add_part(Part(fem_name, fem=fem))
    return


def med_to_fem(fem_file, fem_name):
    from ada import Node

    f = h5py.File(fem_file, "r")

    # Mesh ensemble
    mesh_ensemble = f["ENS_MAA"]
    meshes = mesh_ensemble.keys()
    if len(meshes) != 1:
        raise ValueError("Must only contain exactly 1 mesh, found {}.".format(len(meshes)))
    mesh_name = list(meshes)[0]
    mesh = mesh_ensemble[mesh_name]

    dim = mesh.attrs["ESP"]

    fem_name = fem_name if fem_name is not None else mesh_name

    # Initialize FEM object
    fem = FEM(fem_name)

    # Possible time-stepping
    if "NOE" not in mesh:
        # One needs NOE (node) and MAI (French maillage, meshing) data. If they
        # are not available in the mesh, check for time-steppings.
        time_step = mesh.keys()
        if len(time_step) != 1:
            raise ValueError(f"Must only contain exactly 1 time-step, found {len(time_step)}.")
        mesh = mesh[list(time_step)[0]]

    # Points
    pts_dataset = mesh["NOE"]["COO"]
    n_points = pts_dataset.attrs["NBR"]
    points = pts_dataset[()].reshape((n_points, dim), order="F")

    if "NUM" in mesh["NOE"]:
        point_num = list(mesh["NOE"]["NUM"])
    else:
        logging.warning("No node information is found on MED file")
        point_num = np.arange(1, len(points) + 1)

    fem.nodes = Nodes([Node(p, point_num[i], parent=fem) for i, p in enumerate(points)], parent=fem)

    # Point tags
    tags = None
    if "FAM" in mesh["NOE"]:
        tags = mesh["NOE"]["FAM"][()]

    # Information for point tags
    point_tags = {}
    fas = mesh["FAS"] if "FAS" in mesh else f["FAS"][mesh_name]
    if "NOEUD" in fas:
        point_tags = _read_families(fas["NOEUD"])

    point_sets = _point_tags_to_sets(tags, point_tags, fem) if tags is not None else []

    # Information for cell tags
    cell_tags = {}
    if "ELEME" in fas:
        cell_tags = _read_families(fas["ELEME"])

    # CellBlock
    cell_types = []
    med_cells = mesh["MAI"]

    elements = []
    element_sets = dict()
    for med_cell_type, med_cell_type_group in med_cells.items():
        if med_cell_type == "PO1":
            logging.warning("Point elements are still not supported")
            continue

        cell_type = med_to_ada_type(med_cell_type)
        cell_types.append(cell_type)

        nod = med_cell_type_group["NOD"]
        n_cells = nod.attrs["NBR"]
        nodes_in = nod[()].reshape(n_cells, -1, order="F")

        if "NUM" in med_cell_type_group.keys():
            num = list(med_cell_type_group["NUM"])
        else:
            num = np.arange(0, len(nodes_in))

        element_block = [
            Elem(num[i], [fem.nodes.from_id(e) for e in c], cell_type, parent=fem) for i, c in enumerate(nodes_in)
        ]
        elements += element_block
        # Cell tags
        if "FAM" in med_cell_type_group:
            cell_data = med_cell_type_group["FAM"][()]
            cell_type_sets = _cell_tag_to_set(cell_data, cell_tags)
            for key, val in cell_type_sets.items():
                if key not in element_sets.keys():
                    element_sets[key] = []
                element_sets[key] += [element_block[i] for i in val]

    fem.elements = FemElements(elements, fem)

    elsets = []
    for name, values in element_sets.items():
        elsets.append(FemSet(name, values, "elset", parent=fem))

    fem.sets = FemSets(elsets + point_sets, fem_obj=fem)
    return fem


def _cell_tag_to_set(cell_data_array, cell_tags):
    """
    For a single element type convert tag data into set data

    :param cell_data_array:
    :param cell_tags:
    :return: Cell Sets dictionary
    """
    cell_sets = dict()
    shared_sets = []
    for tag_id, tag_names in cell_tags.items():
        if len(tag_names) > 1:
            for v in tag_names:
                res = np.where(cell_data_array == tag_id)[0]
                if len(res) > 0:
                    shared_sets.append((v, res))
        else:
            tag_name = tag_names[0]
            res = np.where(cell_data_array == tag_id)[0]
            if len(res) > 0:
                cell_sets[tag_name] = res

    for v, s in shared_sets:
        if v in cell_sets.keys():
            cell_sets[v] = np.concatenate([cell_sets[v], s])
        else:
            cell_sets[v] = s

    return cell_sets


def _point_tags_to_sets(tags, point_tags, fem):
    """

    :param tags:
    :param point_tags:
    :return: Point sets dictionary
    """
    point_sets = dict()
    shared_sets = []
    for key, val in point_tags.items():
        if len(val) > 1:
            for set_name in val:
                shared_sets.append((set_name, np.where(tags == key)[0]))
        else:
            point_sets[val[0]] = np.where(tags == key)[0]

    for set_name, s in shared_sets:
        point_sets[set_name] = np.concatenate([point_sets[set_name], s])

    nsets = [FemSet(pn, [fem.nodes.from_id(i + 1) for i in ps], "nset", parent=fem) for pn, ps in point_sets.items()]
    return nsets


def _read_nodal_data(med_data, profiles):
    profile = med_data["NOE"].attrs["PFL"]
    data_profile = med_data["NOE"][profile]
    n_points = data_profile.attrs["NBR"]
    if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
        values = data_profile["CO"][()].reshape(n_points, -1, order="F")
    else:
        n_data = profiles[profile].attrs["NBR"]
        index_profile = profiles[profile]["PFL"][()] - 1
        values_profile = data_profile["CO"][()].reshape(n_data, -1, order="F")
        values = np.full((n_points, values_profile.shape[1]), np.nan)
        values[index_profile] = values_profile
    if values.shape[-1] == 1:  # cut off for scalars
        values = values[:, 0]
    return values


def _read_cell_data(med_data, profiles):
    profile = med_data.attrs["PFL"]
    data_profile = med_data[profile]
    n_cells = data_profile.attrs["NBR"]
    n_gauss_points = data_profile.attrs["NGA"]
    if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
        values = data_profile["CO"][()].reshape(n_cells, n_gauss_points, -1, order="F")
    else:
        n_data = profiles[profile].attrs["NBR"]
        index_profile = profiles[profile]["PFL"][()] - 1
        values_profile = data_profile["CO"][()].reshape(n_data, n_gauss_points, -1, order="F")
        values = np.full((n_cells, values_profile.shape[1], values_profile.shape[2]), np.nan)
        values[index_profile] = values_profile

    # Only 1 data point per cell, shape -> (n_cells, n_components)
    if n_gauss_points == 1:
        values = values[:, 0, :]
        if values.shape[-1] == 1:  # cut off for scalars
            values = values[:, 0]
    return values


def _read_families(fas_data):
    families = {}
    for _, node_set in fas_data.items():
        set_id = node_set.attrs["NUM"]  # unique set id
        n_subsets = node_set["GRO"].attrs["NBR"]  # number of subsets
        nom_dataset = node_set["GRO"]["NOM"][()]  # (n_subsets, 80) of int8
        name = [None] * n_subsets
        for i in range(n_subsets):
            name[i] = "".join([chr(x) for x in nom_dataset[i]]).strip().rstrip("\x00")
        families[set_id] = name
    return families



top_level_fem_str = """IDENT     1.00000000E+00  1.00000000E+00  3.00000000E+00  0.00000000E+00
DATE      1.00000000E+00  0.00000000E+00  4.00000000E+00  7.20000000E+01
        DATE:     {date_str}         TIME:          {clock_str}
        PROGRAM:  ADA python          VERSION:       Not Applicable
        COMPUTER: X86 Windows         INSTALLATION:
        USER:     {user}            ACCOUNT:     \n"""

sestra_header_inp_str = """HEAD
COMM
COMM    Created by: ADA
COMM
COMM    Date : {date_str}   Time : {clock_str}   User : {user}
COMM"""

sestra_eig_inp_str = """
COMM  CHCK ANTP MSUM MOLO STIF RTOP LBCK      PILE CSING     SINGM
CMAS    0.   2.   1.   1.   0.   0.   0.        0.
COMM
COMM                 WCOR THCK
ELOP                   1.   0.
COMM
COMM  ITYP
ITOP    {supnr}.
COMM
COMM  PREFIX
INAM  {name}
COMM
COMM  PREFIX FORMAT
LNAM  {name} UNFORMATTED
COMM
COMM  PREFIX FORMAT
RNAM  {name} NORSAM
COMM
COMM  SEL1 SEL2 SEL3 SEL4 SEL5 SEL6 SEL7 SEL8
RSEL    1.   0.   0.   0.   0.   0.   1.   0.
COMM
COMM  RTRA
RETR    3.
COMM
COMM  ENR                                                              SHIFT
EIGA   {modes}.                                                         0.
COMM
COMM  SELT
IDTY    1.
COMM
COMM  IMAS IDAM ISST
DYMA    1.   0.   0.
Z"""

sestra_static_inp_str = """
COMM  CHCK ANTP MSUM MOLO STIF RTOP LBCK      PILE CSING     SINGM
CMAS    0.   1.   1.   0.   0.   0.   0.        0.
COMM
COMM            ORDR                          CACH MFRWORK
SOLM              0.                            0.        0.
COMM
COMM                 WCOR THCK
ELOP                   1.   0.
COMM
COMM  ITYP
ITOP   {supnr}.
COMM
COMM  PREFIX
INAM  {name}
COMM
COMM  PREFIX FORMAT
LNAM  {name} UNFORMATTED
COMM
COMM  PREFIX FORMAT
RNAM  {name} NORSAM
COMM
COMM  SEL1 SEL2 SEL3 SEL4 SEL5 SEL6 SEL7 SEL8
RSEL    1.   0.   0.   0.   0.   0.   1.   0.
COMM
COMM  RTRA
RETR    3.
COMM
COMM  SELT
IDTY   {supnr}.
COMM
COMM  IMAS IDAM ISST
DYMA    1.   0.   0.
Z"""



from typing import Callable, List, Tuple

import numpy as np

from ada import Assembly, Beam, Material, Node, Part, Pipe, Plate, PrimSphere, Section
from ada.concepts.transforms import Placement
from ada.core.clash_check import penetration_check
from ada.core.constants import X, Y, Z
from ada.core.utils import Counter
from ada.fem import Bc, FemSet
from ada.materials.metals import CarbonSteel

bm_name = Counter(1, "bm")
pl_name = Counter(1, "pl")
floor_name = Counter(1, "floor")


class ReinforcedFloor(Part):
    def __init__(
        self,
        name,
        points: List[tuple],
        pl_thick: float,
        spacing=0.4,
        s_type="HP140x8",
        stringer_dir="X",
        use3dnodes=True,
        **kwargs,
    ):
        super(ReinforcedFloor, self).__init__(name, **kwargs)
        plate = Plate(
            name + "_pl",
            points,
            pl_thick,
            placement=self.placement,
            use3dnodes=use3dnodes,
        )
        self.add_plate(plate)

        # Calculate number of stringers
        bbox = plate.bbox
        xmin, xmax = bbox[0]
        ymin, ymax = bbox[1]

        if stringer_dir == "Y":
            snum = int((xmax - xmin) / spacing) - 1
        else:
            snum = int((ymax - ymin) / spacing) - 1

        z = plate.poly.placement.origin[2]

        tot_spacing = snum * spacing
        diff = xmax - xmin - tot_spacing
        x = diff / 2
        y = diff / 2
        for i in range(0, snum + 1):
            if stringer_dir == "Y":
                p1 = (x, ymin, z)
                p2 = (x, ymax, z)
                x += spacing
            else:
                p1 = (xmin, y, z)
                p2 = (xmax, y, z)
                y += spacing
            self.add_beam(Beam(next(bm_name), p1, p2, sec=s_type))


class SimpleStru(Part):
    class Params:
        w = None
        l = None
        h = None

    def __init__(
        self,
        name,
        w=5,
        l=5,
        h=3,
        gsec="IPE200",
        csec="HEB200",
        pl_thick=10e-3,
        placement=Placement(),
    ):
        super(SimpleStru, self).__init__(name, placement=placement)
        self.Params.w = w
        self.Params.h = h
        self.Params.l = l

        # Define the 5 corner points of each storey
        c1, c2, c3, c4 = self.c1, self.c2, self.c3, self.c4

        # Define the relationship of corners that make up the 4 support beams
        beams = [(c1, c2), (c2, c3), (c3, c4), (c4, c1)]

        z0 = 0
        z1 = h
        sec = Section(gsec, from_str=gsec, parent=self)
        self._elevations = [z0, z1]
        for elev in self._elevations:
            for p1, p2 in beams:
                self.add_beam(Beam(next(bm_name), n1=p1(elev), n2=p2(elev), sec=sec, jusl=Beam.JUSL_TYPES.TOS))
            points = [c1(elev), c2(elev), c3(elev), c4(elev)]
            p = self.add_part(ReinforcedFloor(next(floor_name), points, pl_thick))
            self.add_set("floors", [p])

        # Columns
        columns = [(c1(z0), c1(h)), (c2(z0), c2(h)), (c3(z0), c3(h)), (c4(z0), c4(h))]
        for p1, p2 in columns:
            bm = self.add_beam(Beam(next(bm_name), n1=p1, n2=p2, sec=csec))
            self.add_set("columns", [bm])

    def c1(self, z) -> tuple:
        return 0, 0, z

    def c2(self, z) -> tuple:
        return self.Params.w, 0, z

    def c3(self, z) -> tuple:
        return self.Params.w, self.Params.l, z

    def c4(self, z) -> tuple:
        return 0, self.Params.l, z

    def add_bcs(self):
        funcs: List[Callable] = [self.c1, self.c2, self.c3, self.c4]
        fem_set_btn = self.fem.add_set(FemSet("fix", [], FemSet.TYPES.NSET))
        nodes: List[Node] = []
        for bc_loc in funcs:
            location = self.placement.origin + bc_loc(self._elevations[0])
            nodes += self.fem.nodes.get_by_volume(location)
        fem_set_btn.add_members(nodes)
        self.fem.add_bc(Bc("bc_fix", fem_set_btn, [1, 2, 3]))


def make_it_complex():
    pm = SimpleStru("ParametricModel")
    a = Assembly("ParametricSite") / pm

    elev = pm.Params.h - 0.4
    offset_y = 0.4
    pipe1 = Pipe(
        "Pipe1",
        [
            (0, offset_y, elev),
            (pm.Params.w + 0.4, offset_y, elev),
            (pm.Params.w + 0.4, pm.Params.l + 0.4, elev),
            (pm.Params.w + 0.4, pm.Params.l + 0.4, 0.4),
            (0, pm.Params.l + 0.4, 0.4),
        ],
        Section("PSec1", "PIPE", r=0.1, wt=10e-3),
    )

    pipe2 = Pipe(
        "Pipe2",
        [
            (0.5, offset_y + 0.5, elev + 1.4),
            (0.5, offset_y + 0.5, elev),
            (0.2 + pm.Params.w, offset_y + 0.5, elev),
            (0.2 + pm.Params.w, pm.Params.l + 0.4, elev),
            (0.2 + pm.Params.w, pm.Params.l + 0.4, 0.6),
            (0, pm.Params.l + 0.4, 0.6),
        ],
        Section("PSec2", "PIPE", r=0.05, wt=5e-3),
    )
    pm.add_part(Part("Piping") / [pipe1, pipe2])
    for p in pm.parts.values():
        if type(p) is ReinforcedFloor:
            penetration_check(p)

    return a


class EquipmentTent(Part):
    def __init__(
        self,
        name,
        mass: float,
        cog: Tuple[float, float, float],
        legs=4,
        height=2,
        width=3,
        length=3,
        sec_str="BG200x200x30x30",
        eq_mat=Material("EqMatSoft", CarbonSteel("S355", E=2.1e9)),
    ):
        """

        :param name:
        :param mass:
        :param cog:
        :param legs: can be either 3 or 4 legs
        :param height:
        :param width:
        :param length: Length is along the Y-axis
        """
        super(EquipmentTent, self).__init__(name=name)

        eq_bm = Counter(1, f"{name}_bm")
        cognp = np.array(cog)
        corner_index = [(-1, -1), (-1, 1), (1, 1), (1, -1)]
        if legs == 3:
            corner_index = [(-1, -1), (1, -1), (0.5, 1)]

        mid_points = []
        btn_points = []

        for sX, sY in corner_index:
            p = cognp + sX * np.array(X) * width + sY * np.array(Y) * length - np.array(Z) * height / 2
            mid_points.append(p)
            p = cognp + sX * np.array(X) * width + sY * np.array(Y) * length - np.array(Z) * height
            btn_points.append(p)

        vertical_legs = []
        for btnp, midp in zip(btn_points, mid_points):
            bm = Beam(next(eq_bm), btnp, midp, sec_str, eq_mat)
            vertical_legs.append(bm)
            self.add_beam(bm)

        horizontal_members = []
        for bs, be in zip(mid_points[:-1], mid_points[1:]):
            bm = Beam(next(eq_bm), bs, be, sec_str, eq_mat)
            horizontal_members.append(bm)
            self.add_beam(bm)

        bm = Beam(next(eq_bm), mid_points[-1], mid_points[0], sec_str, eq_mat)
        horizontal_members.append(bm)
        self.add_beam(bm)

        eq_braces = []
        for midp in mid_points:
            bm = Beam(next(eq_bm), midp, cog, sec_str, eq_mat)
            eq_braces.append(bm)
            self.add_beam(bm)

        self.add_set("vertical_members", vertical_legs)
        self.add_set("horizontal_members", horizontal_members)
        self.add_set("braces", eq_braces)

        self.add_shape(PrimSphere(f"{name}_cog", cog, radius=(width + length) / 6, mass=mass))
        self._centre_of_gravity = cog
        self._mass = mass

    def _on_import(self):
        print("Evaluate footing based on existing structure")



import logging
from typing import List, Union

import numpy as np

from .common import Amplitude, Csys, FemBase
from .constants import GRAVITY
from .exceptions.model_definition import UnsupportedLoadType
from .sets import FemSet
from .surfaces import Surface


class LoadTypes:
    GRAVITY = "gravity"
    ACC = "acc"
    ACC_ROT = "acc_rot"
    FORCE = "force"
    FORCE_SET = "force_set"
    MASS = "mass"
    PRESSURE = "pressure"

    all = [GRAVITY, ACC, ACC_ROT, FORCE, FORCE_SET, MASS]


class Load(FemBase):
    """


    :param load_type: Type of loads. See Load.TYPES for allowable load types.
    :param magnitude: Magnitude of load
    :param name: (Required in the event of a point load being applied)
    :param fem_set: Set reference (Required in the event of a point load being applied)
    :param dof: Degrees of freedom (Required in the event of a point load being applied)
    :param follower_force: Should follower force be accounted for
    :param amplitude: Attach an amplitude object to the load
    :param accr_origin: Origin of a rotational Acceleration field (necessary for load_type='acc_rot').
    :type parent: ada.FEM
    """

    TYPES = LoadTypes

    def __init__(
        self,
        name: str,
        load_type: str,
        magnitude: float,
        fem_set: FemSet = None,
        dof: Union[int, List[int]] = None,
        amplitude: Amplitude = None,
        follower_force=False,
        acc_vector=None,
        accr_origin=None,
        accr_rot_axis=None,
        csys: Csys = None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._type = load_type
        self._magnitude = magnitude
        self._fem_set = fem_set

        if type(dof) is int:
            dofs = [None, None, None, None, None, None]
            dofs[dof - 1] = 1
            self._dof = dofs
        else:
            self._dof = dof

        self._amplitude = amplitude
        self._follower_force = follower_force
        self._acc_vector = acc_vector
        self._accr_origin = accr_origin
        self._accr_rot_axis = accr_rot_axis
        self._csys = csys

        if self.type == LoadTypes.FORCE:
            if self._dof is None or self._fem_set is None or self._name is None:
                raise Exception("self._dofs and nid (Node id) and name needs to be set in order to use point loads")

            if len(self._dof) != 6:
                raise Exception(
                    "You need to include all 6 dofs even though forces are not applied in all 6 dofs. "
                    "Use None or 0.0 for the dofs not applied with forces"
                )

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if value.lower() not in LoadTypes.all:
            raise UnsupportedLoadType(f'Load type "{value}" is not among supported load types: "{LoadTypes.all}"')
        self._type = value

    @property
    def dof(self):
        if self._dof is None and self.type is GRAVITY:
            self._dof = [0, 0, 1] if self._dof is None else self._dof
        return self._dof

    @property
    def forces(self):
        if self.type not in (LoadTypes.FORCE,):
            return None

        return [x * self.magnitude if x is not None else 0.0 for x in self.dof]

    @property
    def forces_global(self):
        if self.type not in (LoadTypes.FORCE,):
            return None
        if self.csys is None:
            return [x * self.magnitude for x in self.dof]
        else:
            csys = self.csys
            if csys.coords is None:
                logging.error("Calculating global forces without COORDS is not yet supported")
                return None

            from ada.core.utils import rotation_matrix_csys_rotate

            destination_csys = [(1, 0, 0), (0, 1, 0)]
            rmat = rotation_matrix_csys_rotate(csys.coords, destination_csys)
            res = np.concatenate([np.dot(rmat, np.array(self.dof[:3])), np.dot(rmat, np.array(self.dof[3:]))])
            return [x * self.magnitude for x in res]

    @property
    def amplitude(self):
        return self._amplitude

    @property
    def magnitude(self):
        return self._magnitude

    @property
    def fem_set(self):
        return self._fem_set

    @property
    def follower_force(self):
        return self._follower_force

    @property
    def acc_vector(self):
        if self.type not in (LoadTypes.ACC, LoadTypes.ACC_ROT):
            raise ValueError(f'Acceleration vector only applies for type "acc". Not "{self.type}"')

        dir_error = "If acc_vector is not specified, you must pass dof=[int] (int 1-3) for the acc field"

        if self._acc_vector is not None:
            return self._acc_vector

        dofs = [x for x in self.dof if x is not None]
        if len(dofs) != 1:
            raise ValueError(dir_error)

        acc_magnitude = dofs[0]
        acc_dir = int(self.dof.index(acc_magnitude) + 1)
        if 1 > acc_dir > 3:
            raise ValueError(f"Acceleration vector works only on the translational DOFs. DOF={acc_dir} was passed in")

        if acc_dir == 1:
            dvec = 1, 0, 0
        elif acc_dir == 2:
            dvec = 0, 1, 0
        else:
            dvec = 0, 0, 1

        return tuple([float(self._magnitude * d) if d != 0 else 0.0 for d in dvec])

    @property
    def acc_rot_origin(self):
        return self._accr_origin

    @acc_rot_origin.setter
    def acc_rot_origin(self, value):
        self._accr_origin = value

    @property
    def acc_rot_axis(self):
        return self._accr_rot_axis

    @property
    def csys(self) -> Csys:
        return self._csys

    def __repr__(self):
        forc_str = ",".join(f"{f:.6E}" for f in self.forces)
        return f"Load({self.name}, {self.type}, [{forc_str}])"


class PressureDistribution:
    UNIFORM = "uniform"
    TOTAL_FORCE = "total_force"


class LoadPressure(Load):
    P_DIST_TYPES = PressureDistribution

    def __init__(self, name: str, magnitude: float, surface: Surface, distribution=P_DIST_TYPES.UNIFORM):
        super(LoadPressure, self).__init__(name, LoadTypes.PRESSURE, magnitude)
        self._surface = surface
        self._distribution = distribution

    @property
    def surface(self) -> Surface:
        return self._surface

    @property
    def distribution(self) -> str:
        return self._distribution


class LoadCase(FemBase):
    def __init__(
        self,
        name,
        comment,
        loads=None,
        mass=None,
        lcsys=None,
        metadata=None,
        parent=None,
    ):
        super().__init__(name, metadata, parent)
        self._comment = comment
        self._loads = loads
        self._mass = mass
        self._lcsys = lcsys

    @property
    def loads(self):
        return self._loads

    @property
    def mass(self):
        return self._mass

    @property
    def comment(self):
        return self._comment

    @property
    def csys(self):
        return (1, 0, 0), (0, 1, 0), (0, 0, 1) if self._lcsys is None else self._lcsys

    def __repr__(self):
        return f"LC({self.name}, {self.comment})"



import logging
import traceback
from itertools import chain
from typing import Iterable, List

import numpy as np

from ada.concepts.levels import Part
from ada.concepts.piping import PipeSegStraight
from ada.concepts.points import Node
from ada.concepts.primitives import PrimCyl
from ada.concepts.structural import Beam, Plate

from .utils import Counter, intersect_calc, is_parallel, vector_length


def basic_intersect(bm: Beam, margins, all_parts: [Part]):
    if bm.section.type == "gensec":
        return bm, []
    try:
        vol = bm.bbox
    except ValueError as e:
        logging.error(f"Intersect bbox skipped: {e}\n{traceback.format_exc()}")
        return None
    vol_in = [x for x in zip(vol[0], vol[1])]
    beams = filter(
        lambda x: x != bm,
        chain.from_iterable([p.beams.get_beams_within_volume(vol_in, margins=margins) for p in all_parts]),
    )
    return bm, beams


def beam_cross_check(bm1: Beam, bm2: Beam, outofplane_tol=0.1):
    """Calculate intersection of beams and return point, s, t"""
    p_check = is_parallel
    i_check = intersect_calc
    v_len = vector_length
    a = bm1.n1.p
    b = bm1.n2.p
    c = bm2.n1.p
    d = bm2.n2.p

    ab = b - a
    cd = d - c

    s, t = i_check(a, c, ab, cd)

    ab_ = a + s * ab
    cd_ = c + t * cd

    if p_check(ab, cd):
        logging.debug(f"beams {bm1} {bm2} are parallel")
        return None

    if v_len(ab_ - cd_) > outofplane_tol:
        logging.debug("The two lines do not intersect within given tolerances")
        return None

    return ab_, s, t


def are_beams_connected(bm1: Beam, beams: List[Beam], out_of_plane_tol, point_tol, nodes, nmap) -> None:
    for bm2 in beams:
        if bm1 == bm2:
            continue
        res = beam_cross_check(bm1, bm2, out_of_plane_tol)
        if res is None:
            continue
        point, s, t = res
        t_len = (abs(t) - 1) * bm2.length
        s_len = (abs(s) - 1) * bm1.length
        if t_len > bm2.length / 2 or s_len > bm1.length / 2:
            continue
        if point is not None:
            new_node = Node(point)
            n = nodes.add(new_node, point_tol=point_tol)
            if n not in nmap.keys():
                nmap[n] = [bm1]
            if bm1 not in nmap[n]:
                nmap[n].append(bm1)
            if bm2 not in nmap[n]:
                nmap[n].append(bm2)


def are_plates_touching(pl1: Plate, pl2: Plate, tol=1e-3):
    """Check if two plates are within tolerance of each other"""
    from ..occ.utils import compute_minimal_distance_between_shapes

    dss = compute_minimal_distance_between_shapes(pl1.solid, pl2.solid)
    if dss.Value() <= tol:
        return dss
    else:
        return None


def filter_away_beams_along_plate_edges(pl: Plate, beams: Iterable[Beam]) -> List[Beam]:
    corners = [n for n in pl.poly.points3d]

    # filter away all beams with both ends on any of corner points of the plate
    beams_not_along_plate_edge = []

    for bm in beams:
        t1 = tuple(bm.n1.p)
        t2 = tuple(bm.n2.p)

        if t1 in corners:
            cindex = corners.index(t1)

            nextp = corners[0] if cindex == len(corners) - 1 else corners[cindex + 1]
            prevp = corners[cindex - 1]

            if t2 == nextp or t2 == prevp:
                continue
        beams_not_along_plate_edge.append(bm)

    return beams_not_along_plate_edge


def filter_beams_along_plate_edges(pl: Plate, beams: Iterable[Beam]):
    from .utils import is_clockwise, is_on_line

    corners = [n for n in pl.poly.points3d]
    corners += [corners[0]]
    if is_clockwise(corners):
        corners.reverse()

    # Evalute Corner Points
    crossing_beams = []
    for s, e in zip(corners[:-1], corners[1:]):
        li = (s, e)
        res = [x for x in map(is_on_line, [(li, bm) for bm in beams]) if x is not None]
        crossing_beams += filter(lambda x: x not in crossing_beams, [r[1] for r in res])

    return crossing_beams


def find_beams_connected_to_plate(pl: Plate, beams: List[Beam]) -> List[Beam]:
    """Return all beams with their midpoints inside a specified plate for a given list of beams"""
    from ada.concepts.containers import Nodes

    bbox = list(zip(*pl.bbox))
    nid = Counter(1)
    nodes = Nodes([Node((bm.n2.p + bm.n1.p) / 2, next(nid), refs=[bm]) for bm in beams])
    res = nodes.get_by_volume(bbox[0], bbox[1])

    all_beams_within = list(chain.from_iterable([r.refs for r in res]))
    return all_beams_within


def penetration_check(part: Part):
    a = part.get_assembly()
    cog = part.nodes.vol_cog
    normal = part.placement.zdir
    for p in a.get_all_subparts():
        for pipe in p.pipes:
            for segment in pipe.segments:
                if type(segment) is PipeSegStraight:
                    assert isinstance(segment, PipeSegStraight)
                    p1, p2 = segment.p1, segment.p2
                    v1 = (p1.p - cog) * normal
                    v2 = (p2.p - cog) * normal
                    if np.dot(v1, v2) < 0:
                        part.add_penetration(
                            PrimCyl(f"{p.name}_{pipe.name}_{segment.name}_pen", p1.p, p2.p, pipe.section.r + 0.1)
                        )



import logging
import os
import pathlib
from typing import List, Union

from ada.fem import StepEigen
from ada.fem.concepts.eigenvalue import EigenDataSummary, EigenMode
from ada.fem.formats.utils import DatFormatReader
from ada.fem.results import Results


def get_eigen_data(dat_file: Union[str, os.PathLike]) -> EigenDataSummary:
    dtr = DatFormatReader()

    re_compiled = dtr.compile_ff_re([int] + [float] * 5)
    re_compiled_2 = dtr.compile_ff_re([int] + [float] * 6)

    eig_str = "eigenvalueoutput"
    part_str = "participationfactors"
    eff_modal = "effectivemass"

    eig_res = dtr.read_data_lines(dat_file, re_compiled, eig_str, part_str, split_data=True)
    part_res = dtr.read_data_lines(dat_file, re_compiled_2, part_str, eff_modal, split_data=True)
    modalmass = dtr.read_data_lines(dat_file, re_compiled_2, eff_modal, split_data=True)

    eigen_modes: List[EigenMode] = []

    dof_base = ["x", "y", "z", "rx", "ry", "rz"]
    part_factor_names = ["p" + x for x in dof_base]
    eff_mass_names = ["ef" + x for x in dof_base]

    # Note! participation factors and effective modal mass are each deconstructed into 6 degrees of freedom
    for eig, part, modal in zip(eig_res, part_res, modalmass):
        mode, eig_value, freq_rad, freq_cycl, gen_mass, composite_modal_damping = eig
        eig_output = dict(
            eigenvalue=eig_value,
            f_rad=freq_rad,
            f_hz=freq_cycl,
        )
        participation_data = {pn: p for pn, p in zip(part_factor_names, part[1:])}
        eff_mass_data = {pn: p for pn, p in zip(eff_mass_names, part[1:])}
        eigen_modes.append(EigenMode(no=mode, **eig_output, **participation_data, **eff_mass_data))

    return EigenDataSummary(eigen_modes)


def read_abaqus_results(results: Results, file_ref: pathlib.Path, overwrite):
    dat_file = file_ref.with_suffix(".dat")
    if results.assembly is not None and results.assembly.fem.steps[0] == StepEigen:
        # TODO: Figure out if it is worthwhile adding support for reading step information or if it should be explicitly
        #   stated
        pass

    if dat_file.exists():
        results.eigen_mode_data = get_eigen_data(dat_file)

    logging.error("Result mesh data extraction is not supported for abaqus")
    return None



import logging

from ada.config import Settings as _Settings
from ada.ifc.utils import create_guid


class Backend:
    def __init__(self, name, guid=None, metadata=None, units="m", parent=None, ifc_settings=None, ifc_elem=None):
        self.name = name
        self.parent = parent
        self._ifc_settings = ifc_settings
        self.guid = create_guid() if guid is None else guid
        units = units.lower()
        if units not in _Settings.valid_units:
            raise ValueError(f'Unit type "{units}"')
        self._units = units
        self._metadata = metadata if metadata is not None else dict(props=dict())
        self._ifc_elem = ifc_elem
        # TODO: Currently not able to keep and edit imported ifc_elem objects
        self._ifc_elem = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if _Settings.convert_bad_names:
            logging.debug("Converting bad name")
            value = value.replace("/", "_").replace("=", "")
            if str.isnumeric(value[0]):
                value = "ADA_" + value

        if "/" in value:
            logging.debug(f'Character "/" found in {value}')

        self._name = value.strip()

    @property
    def guid(self):
        return self._guid

    @guid.setter
    def guid(self, value):
        if value is None:
            raise ValueError("guid cannot be None")
        self._guid = value

    @property
    def parent(self):
        """

        :return:
        :rtype: ada.Part
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def metadata(self):
        return self._metadata

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        raise NotImplementedError("Assigning units is not yet represented for this object")

    @property
    def ifc_settings(self):
        if self._ifc_settings is None:
            from ada.ifc.utils import default_settings

            self._ifc_settings = default_settings()
        return self._ifc_settings

    @ifc_settings.setter
    def ifc_settings(self, value):
        self._ifc_settings = value

    def get_ifc_elem(self):
        if self._ifc_elem is None:
            self._ifc_elem = self._generate_ifc_elem()
        return self._ifc_elem

    def get_assembly(self):
        """:rtype: ada.Assembly"""
        from ada import Assembly

        parent = self
        still_looking = True
        max_levels = 10
        a = 0
        while still_looking is True:
            if issubclass(type(parent), Assembly) is True:
                still_looking = False
            if parent.parent is None:
                break
            a += 1
            parent = parent.parent
            if a > max_levels:
                logging.info(f"Max levels reached. {self} is highest level element")
                break
        return parent

    def get_ancestors(self):
        ancestry = [self]
        current = self
        while current.parent is not None:
            ancestry.append(current.parent)
            current = current.parent
        return ancestry

    def _generate_ifc_elem(self):
        raise NotImplementedError("")

    def remove(self):
        """Remove this element/part from assembly/part"""
        from ada import Beam, Part, Shape

        if self.parent is None:
            logging.error(f"Unable to delete {self.name} as it does not have a parent")
            return

        # if self._ifc_elem is not None:
        #     a = self.parent.get_assembly()
        # f = a.ifc_file
        # This returns results in a failure error
        # f.remove(self.ifc_elem)

        if type(self) is Part:
            self.parent.parts.pop(self.name)
        elif issubclass(type(self), Shape):
            self.parent.shapes.pop(self.parent.shapes.index(self))
        elif type(self) is Beam:
            self.parent.beams.remove(self)
        else:
            raise NotImplementedError()



from ada.ifc.utils import ifc_p


def create_reference_subrep(f, global_axes):
    model_rep = f.create_entity("IfcGeometricRepresentationContext", None, "Model", 3, 1.0e-05, global_axes, None)
    body_sub_rep = f.create_entity(
        "IfcGeometricRepresentationSubContext",
        "Body",
        "Model",
        None,
        None,
        None,
        None,
        model_rep,
        None,
        "MODEL_VIEW",
        None,
    )
    ref_sub_rep = f.create_entity(
        "IfcGeometricRepresentationSubContext",
        "Reference",
        "Model",
        None,
        None,
        None,
        None,
        model_rep,
        None,
        "GRAPH_VIEW",
        None,
    )

    return {"model": model_rep, "body": body_sub_rep, "reference": ref_sub_rep}


def ifc_vertex(p, f):
    return f.create_entity("IfcVertexPoint", ifc_p(f, p))



import numpy as np
from IPython.display import display
from ipywidgets import HBox, VBox

from .femviz import get_faces_from_fem, magnitude
from .renderer_pythreejs import MyRenderer
from .threejs_utils import faces_to_mesh


def render_mesh(vertices, faces, colors):
    """
    Renders

    :param vertices:
    :param faces:
    :param colors:
    :return:
    """

    mesh = faces_to_mesh("faces", vertices, faces, colors)

    renderer = MyRenderer()
    renderer._displayed_pickable_objects.add(mesh)
    renderer.build_display()
    display(HBox([VBox([HBox(renderer.controls), renderer.renderer]), renderer.html]))


def viz_meshio_mesh(fem, mesh, data_type):
    """

    :param fem:
    :type fem: ada.fem.FEM
    :param mesh:
    :type mesh: meshio.
    :param data_type:
    :type data_type:
    :return:
    :rtype:
    """
    u = np.asarray(mesh.point_data[data_type], dtype="float32")
    vertices = np.asarray(mesh.points, dtype="float32")
    faces = np.asarray(get_faces_from_fem(fem), dtype="uint16").ravel()

    res = [magnitude(u_) for u_ in u]
    max_r = max(res)
    res_norm_col = np.asarray([(x / max_r, 0, 0) for x in res], dtype="float32")

    render_mesh(vertices, faces, res_norm_col)



import logging
from dataclasses import dataclass
from typing import List, Tuple, Union

from ada import FEM, Beam, Section
from ada.core.utils import Counter, make_name_fem_ready
from ada.fem import FemSection
from ada.fem.exceptions.element_support import IncompatibleElements
from ada.fem.shapes import ElemType

from .write_utils import write_ff

shid = Counter(1)
bmid = Counter(1)
sec_n = Counter(1, "_V")
concept = Counter(1)
concept_ircon = Counter(1)


@dataclass
class BmSectionStr:
    sec_str: str
    names_str: str
    tdsconc_str: str
    sconcept_str: str
    scon_mesh: str


def sections_str(fem: FEM, thick_map) -> str:

    sec_ids = []
    sec_str = ""
    names_str = ""
    concept_str = ""
    tdsconc_str, sconcept_str, scon_mesh = "", "", ""

    sec_names = []
    for fem_sec in fem.sections:
        if fem_sec.type == ElemType.LINE:
            res = create_line_section(fem_sec, sec_names, sec_ids)
            if res is None:
                continue
            names_str += res.names_str
            sec_str += res.sec_str
            tdsconc_str += res.tdsconc_str
            sconcept_str += res.sconcept_str
            scon_mesh += res.scon_mesh
        elif fem_sec.type == ElemType.SHELL:
            sec_str += create_shell_section_str(fem_sec, thick_map)
        elif fem_sec.type == ElemType.SOLID:
            sec_str += create_solid_section(fem_sec)
        else:
            raise IncompatibleElements(f"Solid element type {fem_sec.type} is not yet supported for writing to Sesam")

    return names_str + sec_str + concept_str + tdsconc_str + sconcept_str + scon_mesh


def create_shell_section_str(fem_sec: FemSection, thick_map) -> str:
    if fem_sec.thickness not in thick_map.keys():
        sh_id = next(shid)
        thick_map[fem_sec.thickness] = sh_id
    else:
        sh_id = thick_map[fem_sec.thickness]
    return write_ff("GELTH", [(sh_id, fem_sec.thickness, 5)])


def create_line_section(fem_sec: FemSection, sec_names: List[str], sec_ids: List[Section]) -> Union[BmSectionStr, None]:
    from .write_bm_profiles import write_bm_section

    sec = fem_sec.section
    if sec in sec_ids:
        logging.info(f'Skipping already included section "{sec}"')
        return None

    sec_ids.append(fem_sec.section)
    secid = next(bmid)
    sec_name = make_name_fem_ready(fem_sec.section.name, no_dot=True)
    if sec_name not in sec_names:
        sec_names.append(sec_name)
    else:
        sec_name += next(sec_n)

    sec.metadata["numid"] = secid

    names_str = write_ff(
        "TDSECT",
        [
            (4, secid, 100 + len(sec_name), 0),
            (sec_name,),
        ],
    )
    sec_str = write_bm_section(sec, secid)
    if fem_sec.refs is not None:
        tdsconc_str, sconcept_str, scon_mesh = create_sconcept_str(fem_sec)
    else:
        tdsconc_str, sconcept_str, scon_mesh = "", "", ""

    return BmSectionStr(
        sec_str=sec_str, names_str=names_str, tdsconc_str=tdsconc_str, sconcept_str=sconcept_str, scon_mesh=scon_mesh
    )


def create_solid_section(fem_sec: FemSection):
    raise IncompatibleElements(f"Solid element type {fem_sec.type} is not yet supported for writing to Sesam")


def create_sconcept_str(fem_sec: FemSection) -> Tuple[str, str, str]:
    sconcept_str = ""
    # Give concept relationship based on inputted values

    beams = [x for x in fem_sec.refs if type(x) is Beam]
    if len(beams) != 1:
        raise ValueError("A FemSection cannot be sourced from multiple beams")
    beam = beams[0]

    fem_sec.metadata["ircon"] = next(concept_ircon)
    bm_name = make_name_fem_ready(beam.name, no_dot=True)
    tdsconc_str = write_ff(
        "TDSCONC",
        [(4, fem_sec.metadata["ircon"], 100 + len(bm_name), 0), (bm_name,)],
    )
    sconcept_str += write_ff("SCONCEPT", [(8, next(concept), 7, 0), (0, 1, 0, 2)])
    sconc_ref = next(concept)
    sconcept_str += write_ff("SCONCEPT", [(5, sconc_ref, 2, 4), (1,)])
    elids: List[tuple] = []
    i = 0

    numel = len(beam.elem_refs)
    elid_bulk = [numel]
    for el in fem_sec.elset.members:
        if i == 3:
            elids.append(tuple(elid_bulk))
            elid_bulk = []
            i = -1
        elid_bulk.append(el.id)
        i += 1
    if len(elid_bulk) != 0:
        elids.append(tuple(elid_bulk))

    mesh_args = [(5 + numel, sconc_ref, 1, 2)] + elids
    scon_mesh = write_ff("SCONMESH", mesh_args)
    return tdsconc_str, sconcept_str, scon_mesh



from typing import Union

import numpy as np
from OCC.Core.TopoDS import TopoDS_Face, TopoDS_Wire
from OCC.Extend.ShapeFactory import make_face, make_wire

from ada.concepts.transforms import Placement
from ada.sections.categories import SectionCat
from ada.sections.concept import SectionProfile

from .utils import make_circle, make_face_w_cutout


def cross_sec_face(sec_profile: SectionProfile, placement: Placement, solid_repre) -> Union[TopoDS_Face, TopoDS_Wire]:

    inner_shape = None

    if sec_profile.sec.type in SectionCat.tubular:
        outer_shape = make_wire([make_circle(placement.origin, placement.zdir, sec_profile.sec.r)])
        inner_shape = make_wire([make_circle(placement.origin, placement.zdir, sec_profile.sec.r - sec_profile.sec.wt)])
    elif sec_profile.sec.type in SectionCat.circular:
        outer_shape = make_wire([make_circle(placement.origin, placement.zdir, sec_profile.sec.r)])
    elif sec_profile.sec.type in SectionCat.general:
        radius = np.sqrt(sec_profile.sec.properties.Ax / np.pi)
        outer_shape = make_wire([make_circle(placement.origin, placement.zdir, radius)])
    else:
        if sec_profile.disconnected is False:
            outer_curve = sec_profile.outer_curve
            inner_curve = sec_profile.inner_curve
            outer_curve.placement = placement
            outer_shape = outer_curve.face
            if inner_curve is not None:
                inner_curve.placement = placement
                inner_shape = inner_curve.wire
        else:
            outer_shape = []
            for curve in sec_profile.outer_curve_disconnected:
                curve.placement = placement
                outer_shape.append(curve.wire)

    if inner_shape is not None and solid_repre is True:
        try:
            face = make_face(outer_shape) if type(outer_shape) is not TopoDS_Face else outer_shape
            shape = make_face_w_cutout(face, inner_shape)
        except TypeError as e:
            raise TypeError(e)
        except AssertionError as e:
            raise AssertionError(e)
    else:
        shape = outer_shape
    if shape is None:
        raise ValueError("Shape cannot be None")
    return shape
